# Copyright 2021-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition for Pipeline."""

import functools
from typing import Callable, Optional

from kfp.components import component_factory
from kfp.components import pipeline_task
from kfp.components import tasks_group
from kfp.components import utils


def pipeline(func: Optional[Callable] = None,
             *,
             name: Optional[str] = None,
             description: Optional[str] = None,
             pipeline_root: Optional[str] = None,
             display_name: Optional[str] = None) -> Callable:
    """Decorator used to construct a pipeline.

    Example
      ::

        @pipeline(
          name='my-pipeline',
          description='My ML Pipeline.'
          pipeline_root='gs://my-bucket/my-output-path'
        )
        def my_pipeline(a: str, b: int):
          ...

    Args:
        func: The Python function that defines a pipeline.
        name: The pipeline name. Defaults to a sanitized version of the
            decorated function name.
        description: A human-readable description of the pipeline.
        pipeline_root: The root directory from which to read input and output
            parameters and artifacts.
        display_name: A human-readable name for the pipeline.
    """
    if func is None:
        return functools.partial(
            pipeline,
            name=name,
            description=description,
            pipeline_root=pipeline_root,
            display_name=display_name,
        )

    if pipeline_root:
        func.pipeline_root = pipeline_root

    return component_factory.create_graph_component_from_func(
        func,
        name=name,
        description=description,
        display_name=display_name,
    )


class Pipeline:
    """A pipeline contains a list of tasks.

    This class is not supposed to be used by pipeline authors since pipeline
    authors can use pipeline functions (decorated with @pipeline) to reference
    their pipelines.
    This class is useful for implementing a compiler. For example, the compiler
    can use the following to get the pipeline object and its tasks:

    Example:
      ::

        with Pipeline() as p:
            pipeline_func(*args_list)

        traverse(p.tasks)

    Attributes:
        name:
        tasks:
        groups:
    """

    # _default_pipeline is set when the compiler runs "with Pipeline()"
    _default_pipeline = None

    @staticmethod
    def get_default_pipeline():
        """Gets the default pipeline."""
        return Pipeline._default_pipeline

    def __init__(self, name: str):
        """Creates a new instance of Pipeline.

        Args:
            name: The name of the pipeline.
        """
        self.name = name
        self.tasks = {}
        # Add the root group.
        self.groups = [
            tasks_group.TasksGroup(
                group_type=tasks_group.TasksGroupType.PIPELINE,
                name=name,
                is_root=True)
        ]
        self._group_id = 0

    def __enter__(self):

        if Pipeline._default_pipeline:
            raise Exception('Nested pipelines are not allowed.')

        Pipeline._default_pipeline = self

        def register_task_and_generate_id(task: pipeline_task.PipelineTask):
            return self.add_task(
                task=task,
                add_to_group=not getattr(task, 'is_exit_handler', False))

        self._old_register_task_handler = (
            pipeline_task.PipelineTask._register_task_handler)
        pipeline_task.PipelineTask._register_task_handler = (
            register_task_and_generate_id)
        return self

    def __exit__(self, *unused_args):

        Pipeline._default_pipeline = None
        pipeline_task.PipelineTask._register_task_handler = (
            self._old_register_task_handler)

    def add_task(
        self,
        task: pipeline_task.PipelineTask,
        add_to_group: bool,
    ) -> str:
        """Adds a new task.

        Args:
            task: A PipelineTask instance.
            add_to_group: Whether add the task into the current group. Expect
                True for all tasks expect for exit handler.

        Returns:
            A unique task name.
        """
        # Sanitizing the task name.
        # Technically this could be delayed to the compilation stage, but string
        # serialization of PipelineChannels make unsanitized names problematic.
        task_name = utils.maybe_rename_for_k8s(task.component_spec.name)
        #If there is an existing task with this name then generate a new name.
        task_name = utils.make_name_unique_by_adding_index(
            task_name, list(self.tasks.keys()), '-')
        if task_name == '':
            task_name = utils.make_name_unique_by_adding_index(
                'task', list(self.tasks.keys()), '-')

        self.tasks[task_name] = task
        if add_to_group:
            task.parent_task_group = self.groups[-1]
            self.groups[-1].tasks.append(task)

        return task_name

    def push_tasks_group(self, group: 'tasks_group.TasksGroup'):
        """Pushes a TasksGroup into the stack.

        Args:
            group: A TasksGroup. Typically it is one of ExitHandler, Condition,
                and ParallelFor.
        """
        self.groups[-1].groups.append(group)
        self.groups.append(group)

    def pop_tasks_group(self):
        """Removes the current TasksGroup from the stack."""
        del self.groups[-1]

    def remove_task_from_groups(self, task: pipeline_task.PipelineTask):
        """Removes a task from the pipeline.

        This is useful for excluding exit handler from the pipeline.
        """
        for group in self.groups:
            group.remove_task_recursive(task)

    def get_next_group_id(self) -> str:
        """Gets the next id for a new group."""
        self._group_id += 1
        return str(self._group_id)



"""The `kfp.components` module contains functions for loading components from
compiled YAML."""
# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'load_component_from_file',
    'load_component_from_url',
    'load_component_from_text',
    'PythonComponent',
    'BaseComponent',
    'ContainerComponent',
    'YamlComponent',
]

from kfp.components.base_component import BaseComponent
from kfp.components.container_component import ContainerComponent
from kfp.components.python_component import PythonComponent
from kfp.components.yaml_component import load_component_from_file
from kfp.components.yaml_component import load_component_from_text
from kfp.components.yaml_component import load_component_from_url
from kfp.components.yaml_component import YamlComponent



# Copyright 2021 Arrikto Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from kubernetes.client import configuration


class TokenCredentialsBase(abc.ABC):

    @abc.abstractmethod
    def refresh_api_key_hook(self, config: configuration.Configuration):
        """Refresh the api key.

        This is a helper function for registering token refresh with swagger
        generated clients.

        All classes that inherit from TokenCredentialsBase must implement this
        method to refresh the credentials.

        Args:
            config (kubernetes.client.configuration.Configuration):
                The configuration object that the client uses.

                The Configuration object of the kubernetes client's is the same
                with kfp_server_api.configuration.Configuration.
        """
        raise NotImplementedError()


def read_token_from_file(path=None):
    """Read a token found in some file."""
    token = None
    with open(path, "r") as f:
        token = f.read().strip()
    return token



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.types.artifact_types."""

import json
import os
import unittest

from absl.testing import parameterized
from kfp.components.types import artifact_types


class ArtifactsTest(parameterized.TestCase):

    def test_complex_metrics(self):
        metrics = artifact_types.ClassificationMetrics()
        metrics.log_roc_data_point(threshold=0.1, tpr=98.2, fpr=96.2)
        metrics.log_roc_data_point(threshold=24.3, tpr=24.5, fpr=98.4)
        metrics.set_confusion_matrix_categories(['dog', 'cat', 'horses'])
        metrics.log_confusion_matrix_row('dog', [2, 6, 0])
        metrics.log_confusion_matrix_cell('cat', 'dog', 3)
        metrics.log_confusion_matrix_cell('horses', 'horses', 3)
        metrics.metadata['test'] = 1.0
        with open(
                os.path.join(
                    os.path.dirname(__file__), 'test_data',
                    'expected_io_types_classification_metrics.json')
        ) as json_file:
            expected_json = json.load(json_file)
            self.assertEqual(expected_json, metrics.metadata)

    def test_complex_metrics_bulk_loading(self):
        metrics = artifact_types.ClassificationMetrics()
        metrics.log_roc_curve(
            fpr=[85.1, 85.1, 85.1],
            tpr=[52.6, 52.6, 52.6],
            threshold=[53.6, 53.6, 53.6])
        metrics.log_confusion_matrix(['dog', 'cat', 'horses'],
                                     [[2, 6, 0], [3, 5, 6], [5, 7, 8]])
        with open(
                os.path.join(
                    os.path.dirname(__file__), 'test_data',
                    'expected_io_types_bulk_load_classification_metrics.json')
        ) as json_file:
            expected_json = json.load(json_file)
            self.assertEqual(expected_json, metrics.metadata)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
import json
import logging
import os
from typing import Any, Callable, Dict, Generator, Iterable, Tuple, Union
from urllib.parse import parse_qs
from urllib.parse import urlparse
from webbrowser import open_new_tab
import wsgiref.simple_server
import wsgiref.util

import google.auth
import google.auth.app_engine
import google.auth.compute_engine.credentials
import google.auth.iam
from google.auth.transport.requests import Request
import google.oauth2.credentials
import google.oauth2.service_account
import requests
import requests_toolbelt.adapters.appengine

IAM_SCOPE = 'https://www.googleapis.com/auth/iam'
OAUTH_TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'
LOCAL_KFP_CREDENTIAL = os.path.expanduser('~/.config/kfp/credentials.json')


def get_gcp_access_token() -> Union[str, None]:
    """Gets GCP access token for the current Application Default Credentials.

    Returns:
        GCP access token or None, if it is not set. For more information, see
    https://cloud.google.com/sdk/gcloud/reference/auth/application-default/print-access-token.
    """
    token = None
    try:
        creds, _ = google.auth.default(
            scopes=['https://www.googleapis.com/auth/cloud-platform'])
        if not creds.valid:
            auth_req = Request()
            creds.refresh(auth_req)
        if creds.valid:
            token = creds.token
    except Exception as e:
        logging.warning('Failed to get GCP access token: %s', e)
    return token


def get_auth_token(client_id: str, other_client_id: str,
                   other_client_secret: str) -> Tuple[Union[str, None], bool]:
    """Gets auth token from default service account or user account.

    Returns:
        Tuple of (ID token or None, if not found, and a boolean
        indicating whether a refresh token has been saved locally)
    """
    is_refresh_token = True
    if os.path.exists(LOCAL_KFP_CREDENTIAL):
        # fetch IAP auth token using the locally stored credentials.
        with open(LOCAL_KFP_CREDENTIAL, 'r') as f:
            credentials = json.load(f)
        if client_id in credentials:
            saved_refresh_token = credentials[client_id].get('refresh_token')
            saved_other_client_id = credentials[client_id].get(
                'other_client_id')
            saved_other_client_secret = credentials[client_id].get(
                'other_client_secret')
            if None not in {
                    saved_refresh_token, saved_other_client_id,
                    saved_other_client_secret
            }:
                return id_token_from_refresh_token(saved_other_client_id,
                                                   saved_other_client_secret,
                                                   saved_refresh_token,
                                                   client_id), is_refresh_token
    if other_client_id is None or other_client_secret is None:
        # fetch IAP auth token: service accounts
        token = get_auth_token_from_sa(client_id)
        is_refresh_token = False
    else:
        # fetch IAP auth token: user account
        # Obtain the ID token for provided Client ID with user accounts.
        # Flow: get authorization code -> exchange for refresh token -> obtain
        # and return ID token
        refresh_token, is_refresh_token = get_refresh_token_from_client_id(
            other_client_id, other_client_secret)
        credentials = {}
        if os.path.exists(LOCAL_KFP_CREDENTIAL):
            with open(LOCAL_KFP_CREDENTIAL, 'r') as f:
                credentials = json.load(f)
        credentials[client_id] = {
            'other_client_id': other_client_id,
            'other_client_secret': other_client_secret
        }

        if is_refresh_token:
            credentials[client_id]['refresh_token'] = refresh_token
        else:
            credentials[client_id]['access_token'] = refresh_token
        # TODO: handle the case when the refresh_token expires, which only
        # happens if the refresh_token is not used once for six months.
        if not os.path.exists(os.path.dirname(LOCAL_KFP_CREDENTIAL)):
            os.makedirs(os.path.dirname(LOCAL_KFP_CREDENTIAL))
        with open(LOCAL_KFP_CREDENTIAL, 'w') as f:
            json.dump(credentials, f)
        token = id_token_from_refresh_token(other_client_id,
                                            other_client_secret, refresh_token,
                                            client_id)
    return token, is_refresh_token


def get_auth_token_from_sa(client_id: str) -> Union[str, None]:
    """Gets auth token from default service account.

    Returns:
        Authorization token or None, if not found.
    """
    service_account_credentials = get_service_account_credentials(client_id)
    if service_account_credentials:
        return get_google_open_id_connect_token(service_account_credentials)
    return None


def get_service_account_credentials(
        client_id: str
) -> Union[google.oauth2.service_account.Credentials, None]:
    """Figure out what environment we're running in and get some preliminary
    information about the service account.

    Args:
        client_id: OAuth client ID.

    Returns:
        OAuth2 credentials or None.
    """
    bootstrap_credentials, _ = google.auth.default(scopes=[IAM_SCOPE])
    if isinstance(bootstrap_credentials, google.oauth2.credentials.Credentials):
        logging.info('Found OAuth2 credentials and skip SA auth.')
        return None
    if isinstance(bootstrap_credentials, google.auth.app_engine.Credentials):
        requests_toolbelt.adapters.appengine.monkeypatch()
    # For service account's using the Compute Engine metadata service,
    # service_account_email isn't available until refresh is called.
    bootstrap_credentials.refresh(Request())
    signer_email = bootstrap_credentials.service_account_email
    if isinstance(bootstrap_credentials,
                  google.auth.compute_engine.credentials.Credentials):
        # Since the Compute Engine metadata service doesn't expose the service
        # account key, we use the IAM signBlob API to sign instead.
        # In order for this to work:
        #
        # 1. Your VM needs the https://www.googleapis.com/auth/iam scope.
        #    You can specify this specific scope when creating a VM
        #    through the API or gcloud. When using Cloud Console,
        #    you'll need to specify the "full access to all Cloud APIs"
        #    scope. A VM's scopes can only be specified at creation time.
        #
        # 2. The VM's default service account needs the "Service Account Actor"
        #    role. This can be found under the "Project" category in Cloud
        #    Console, or roles/iam.serviceAccountActor in gcloud.
        signer = google.auth.iam.Signer(Request(), bootstrap_credentials,
                                        signer_email)
    else:
        # A Signer object can sign a JWT using the service account's key.
        signer = bootstrap_credentials.signer
    # Construct OAuth 2.0 service account credentials using the signer
    # and email acquired from the bootstrap credentials.
    return google.oauth2.service_account.Credentials(
        signer,
        signer_email,
        token_uri=OAUTH_TOKEN_URI,
        additional_claims={'target_audience': client_id})


def get_google_open_id_connect_token(
    service_account_credentials: google.oauth2.service_account.Credentials
) -> str:
    """Gets an OpenID Connect token issued by Google for the service account.

    This function:
      1. Generates a JWT signed with the service account's private key
         containing a special "target_audience" claim.
      2. Sends it to the OAUTH_TOKEN_URI endpoint. Because the JWT in #1
         has a target_audience claim, that endpoint will respond with
         an OpenID Connect token for the service account -- in other words,
         a JWT signed by *Google*. The aud claim in this JWT will be
         set to the value from the target_audience claim in #1.
    For more information, see
    https://developers.google.com/identity/protocols/OAuth2ServiceAccount
    The HTTP/REST example on that page describes the JWT structure and
    demonstrates how to call the token endpoint. (The example on that page
    shows how to get an OAuth2 access token; this code is using a
    modified version of it to get an OpenID Connect token).

    Returns:
        OAuth ID token.
    """
    service_account_jwt = (
        service_account_credentials._make_authorization_grant_assertion())
    request = google.auth.transport.requests.Request()
    body = {
        'assertion': service_account_jwt,
        'grant_type': google.oauth2._client._JWT_GRANT_TYPE,
    }
    token_response = google.oauth2._client._token_endpoint_request(
        request, OAUTH_TOKEN_URI, body)
    return token_response['id_token']


def get_refresh_token_from_client_id(client_id: str,
                                     client_secret: str) -> Tuple[str, bool]:
    """Obtains the ID token for provided Client ID with user accounts.
    Flow: get authorization code -> exchange for refresh token -> obtain and
    return ID token.

    Args:
        client_id: OAuth client ID.
        client_secret: OAuth client secret.
            See https://console.cloud.google.com/apis/credentials.

    Returns:
        Tuple of (OAuth short-lived access token or long-lived refresh token,
        and a boolean indicating whether the returned token is refresh_token)
    """
    auth_code, redirect_uri = get_auth_code(client_id)
    token, is_refresh_token = get_refresh_token_from_code(
        auth_code, client_id, client_secret, redirect_uri)
    return token, is_refresh_token


def get_auth_code(client_id: str) -> Tuple[str, str]:
    """Retrieves authorization token using Loopback flow.

    Args:
        client_id: OAuth client ID. To retrieve it, visit
          https://console.cloud.google.com/apis/credentials.

    Returns:
        A tuple of (authorization token, redirect_uri parameter).

    Raises:
        ValueError: If the provided authorization_response is empty.
    """
    host = 'localhost'
    port = 9901
    redirect_uri = f'http://{host}:{port}'
    auth_url = ('https://accounts.google.com/o/oauth2/v2/auth?'
                f'client_id={client_id}&response_type=code&'
                'scope=openid%20email&access_type=offline&'
                f'redirect_uri={redirect_uri}')
    authorization_response = None
    if ('SSH_CONNECTION' in os.environ) or ('SSH_CLIENT' in os.environ):
        try:
            print((
                'SSH connection detected. Please follow the instructions below.'
                ' Otherwise, press CTRL+C if you are not connected via SSH.'))
            authorization_response = get_auth_response_ssh(host, port, auth_url)
        except KeyboardInterrupt:
            authorization_response = None
            logging.warning('User pressed CTRL+C. Trying to open browser...')
    if authorization_response is None:
        try:
            print(('Using a local web-server. Please follow the instructions '
                   'below. Otherwise, press CTRL+C to cancel and manually '
                   'copy-paste the response URL.'))
            authorization_response = get_auth_response_local(
                host, port, auth_url)
        except KeyboardInterrupt:
            logging.warning('User pressed CTRL+C. See instructions below.')
            authorization_response = get_auth_response_ssh(host, port, auth_url)
        except OSError as err:
            logging.warning(
                ('%s.\n Error occurred while creating a local web-server. '
                 'Possibly http://%s:%i is allocated to '
                 'another process. Falling back to manual mode. '
                 'See instructions below.'), err, host, port)
            authorization_response = get_auth_response_ssh(host, port, auth_url)
    if authorization_response is None:
        raise ValueError(
            'Authorization response URL is empty. This may be caused by '
            f'corrupted or expired credentials in {LOCAL_KFP_CREDENTIAL}'
            '. Try renaming or moving them to another directory before '
            'running again.')
    token = fetch_auth_token_from_response(authorization_response)
    return token, redirect_uri


def get_auth_response_ssh(host: str, port: int, auth_url: str) -> str:
    """Fetches OAuth authorization response URL for remote SSH connection.

    Args:
        host: Hostname in redirect_uri.
        port: Port in redirect_uri.
        auth_url: OAuth request URL.

    Returns:
        A URL containing authorization code.
    """
    print(auth_url)
    return input(
        'Carefully follow these steps: (1) open the URL above in your'
        ' browser, (2) authenticate and copy a url of the response page'
        f' that starts with http://{host}:{port}..., and (3) paste it'
        ' below:\n')


def get_auth_response_local(host: str, port: int,
                            auth_url: str) -> Union[str, None]:
    """Fetches OAuth authorization response URL using a local web-server.

    Args:
        host: Hostname of the server.
        port: Port of the server.
        auth_url: OAuth request URL.

    Returns:
        A URL containing authorization code.
    """
    with get_local_server_app(host, port) as (local_server, wsgi_app):
        open_new_tab(auth_url)
        print(f'Please visit this URL to authorize Kubeflow SDK: {auth_url}')
        print((f'Make sure that http://{host}:{port} is added to Authorized'
               ' redirect URIs for your OAuth 2.0 Client ID. Check it here:'
               ' https://console.cloud.google.com/apis/credentials'))
        local_server.handle_request()
        return wsgi_app.last_request_uri


def get_refresh_token_from_code(auth_code: str, client_id: str,
                                client_secret: str,
                                redirect_uri: str) -> Tuple[str, bool]:
    """Returns refresh or access token from authorization code.

    Args:
        auth_code: OAuth authorization code.
        client_id: OAuth client ID.
        client_secret: OAuth client secret.
        redirect_uri: Redirect uri used to obtain auth_code.

    Returns:
        Tuple of (OAuth short-lived access token or long-lived refresh token,
        and a boolean indicating whether the returned token is refresh_token)

    Raises:
        ValueError: If HTTP request returns
            a requests.exceptions.HTTPError.
    """
    payload = {
        'code': auth_code,
        'client_id': client_id,
        'client_secret': client_secret,
        'redirect_uri': redirect_uri,
        'grant_type': 'authorization_code'
    }
    res = requests.post(OAUTH_TOKEN_URI, data=payload)
    try:
        res.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise ValueError(
            ('Some HTTPErrors are caused by expired credentials in '
             f'{LOCAL_KFP_CREDENTIAL} Try renaming or moving '
             'them to another directory before running again.')) from err
    parsed_res = json.loads(res.text)
    token = parsed_res.get('refresh_token')
    is_refresh_token = True
    if token is None:
        token = parsed_res.get('access_token')
        is_refresh_token = False
    return str(token), is_refresh_token


def id_token_from_refresh_token(client_id: str, client_secret: str,
                                refresh_token: str, audience: str) -> str:
    """Returns ID token from refresh token.

    Args:
        client_id: OAuth client ID.
        client_secret: OAuth client secret.
        refresh_token: OAuth refresh token.
        audience: OAuth audience.

    Returns:
        OAuth ID token.

    Raises:
        ValueError: If HTTP request returns
            a requests.exceptions.HTTPError.
    """
    payload = {
        'client_id': client_id,
        'client_secret': client_secret,
        'refresh_token': refresh_token,
        'grant_type': 'refresh_token',
        'audience': audience
    }
    res = requests.post(OAUTH_TOKEN_URI, data=payload)
    try:
        res.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise ValueError(
            ('Some HTTPErrors are caused by expired credentials in '
             f'{LOCAL_KFP_CREDENTIAL}. Try renaming or moving '
             'them to another directory before running again.')) from err
    return str(json.loads(res.text).get('id_token'))


class RedirectWSGIApp:
    """WSGI app to handle the authorization redirect.

    Stores the request URI and displays the given success message.
    """

    def __init__(self, success_message: str) -> None:
        """
        Args:
            success_message: The message to display in the web browser
                the authorization flow is complete.
        """
        self.last_request_uri = None
        self._success_message = success_message

    def __call__(
        self, environ: Dict[str, Any], start_response: Callable[[str, list],
                                                                Callable[...,
                                                                         None]]
    ) -> Iterable[bytes]:
        """WSGI Callable. Updates environment dictionary with parameters
        required for WSGI and returns it.

        Args:
            environ: The WSGI environment.
            start_response: The WSGI start_response
                callable.

        Returns:
            The response body.
        """
        wsgiref.util.setup_testing_defaults(environ)
        start_response('200 OK',
                       [('Content-type', 'text/plain; charset=utf-8')])
        self.last_request_uri = wsgiref.util.request_uri(environ)
        return [self._success_message.encode('utf-8')]


@contextmanager
def get_local_server_app(
    host: str, port: int
) -> Generator[Tuple[wsgiref.simple_server.WSGIServer, RedirectWSGIApp], None,
               None]:
    """Creates a local web-server for given host and port.

    Args:
        host: Hostname of the server.
        port: Port of the server.

    Returns:
        Tuple of (a local server instance, WSGI app that handles
            the authorization redirect).
    """
    success_message = ('Kubeflow SDK authentication is completed.'
                       ' You may close this window now.')
    wsgi_app = RedirectWSGIApp(success_message)
    wsgiref.simple_server.WSGIServer.allow_reuse_address = False
    local_server = wsgiref.simple_server.make_server(
        host,
        port,
        wsgi_app,
        handler_class=wsgiref.simple_server.WSGIRequestHandler)
    try:
        yield local_server, wsgi_app
    finally:
        local_server.server_close()
        del wsgi_app


def fetch_auth_token_from_response(url: str) -> str:
    """Fetches authorization code for OAuth2.0 Loopback flow.

    Args:
        url: A string containing the response URL.

    Returns:
        An access code.

    Raises:
        KeyError: If no authorization code is found in the provided url.
    """
    parsed_url = urlparse(url)
    parsed_query = parse_qs(parsed_url.query)
    access_code = parsed_query.get('code')
    if access_code is None:
        raise KeyError((
            'Authorization code is missing or empty in the response.'
            ' Please, try again or check '
            'https://www.kubeflow.org/docs/distributions/gke/deploy/oauth-setup'
        ))
    if isinstance(access_code, list):
        access_code = str(access_code.pop(0))
    return access_code



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from kubernetes.client.models import (V1Volume, V1TypedLocalObjectReference,
                                      V1ObjectMeta)

from ._resource_op import ResourceOp
from ._pipeline_param import match_serialized_pipelineparam, sanitize_k8s_name


class VolumeSnapshotOp(ResourceOp):
    """Represents an op which will be translated into a resource template which
    will be creating a VolumeSnapshot.

    TODO(https://github.com/kubeflow/pipelines/issues/4822): Determine the
        stability level of this feature.

    Args:
      resource_name: A desired name for the VolumeSnapshot which will be
        created
      pvc: The name of the PVC which will be snapshotted
      snapshot_class: The snapshot class to use for the dynamically created
        VolumeSnapshot
      annotations: Annotations to be patched in the VolumeSnapshot
      volume: An instance of V1Volume
      kwargs: See :py:class:`kfp.dsl.ResourceOp`

    Raises:
      ValueError: if k8s_resource is provided along with other arguments
                  if k8s_resource is not a VolumeSnapshot
                  if pvc and volume are None
                  if pvc and volume are not None
                  if volume does not reference a PVC
    """

    def __init__(self,
                 resource_name: str = None,
                 pvc: str = None,
                 snapshot_class: str = None,
                 annotations: Dict[str, str] = None,
                 volume: V1Volume = None,
                 api_version: str = "snapshot.storage.k8s.io/v1alpha1",
                 **kwargs):
        # Add size to output params
        self.attribute_outputs = {"size": "{.status.restoreSize}"}
        # Add default success_condition if None provided
        if "success_condition" not in kwargs:
            kwargs["success_condition"] = "status.readyToUse == true"

        if "k8s_resource" in kwargs:
            if resource_name or pvc or snapshot_class or annotations or volume:
                raise ValueError("You cannot provide k8s_resource along with "
                                 "other arguments.")
            # TODO: Check if is VolumeSnapshot
            super().__init__(**kwargs)
            self.snapshot = V1TypedLocalObjectReference(
                api_group="snapshot.storage.k8s.io",
                kind="VolumeSnapshot",
                name=self.outputs["name"])
            return

        if not (pvc or volume):
            raise ValueError("You must provide a pvc or a volume.")
        elif pvc and volume:
            raise ValueError("You can't provide both pvc and volume.")

        source = None
        deps = []
        if pvc:
            source = V1TypedLocalObjectReference(
                kind="PersistentVolumeClaim", name=pvc)
        else:
            if not hasattr(volume, "persistent_volume_claim"):
                raise ValueError("The volume must be referencing a PVC.")
            if hasattr(volume,
                       "dependent_names"):  #TODO: Replace with type check
                deps = list(volume.dependent_names)
            source = V1TypedLocalObjectReference(
                kind="PersistentVolumeClaim",
                name=volume.persistent_volume_claim.claim_name)

        # Set the k8s_resource
        # TODO: Use VolumeSnapshot
        if not match_serialized_pipelineparam(str(resource_name)):
            resource_name = sanitize_k8s_name(resource_name)
        snapshot_metadata = V1ObjectMeta(
            name="{{workflow.name}}-%s" % resource_name,
            annotations=annotations)
        k8s_resource = {
            "apiVersion": api_version,
            "kind": "VolumeSnapshot",
            "metadata": snapshot_metadata,
            "spec": {
                "source": source
            }
        }
        if snapshot_class:
            k8s_resource["spec"]["snapshotClassName"] = snapshot_class

        super().__init__(k8s_resource=k8s_resource, **kwargs)
        self.dependent_names.extend(deps)
        self.snapshot = V1TypedLocalObjectReference(
            api_group="snapshot.storage.k8s.io",
            kind="VolumeSnapshot",
            name=self.outputs["name"])



# Lint as: python3
# Copyright 2019 The Kubeflow Authors. All Rights Reserved.
#
# Licensed under the Apache License,Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for diagnose_me.gcp."""

from typing import Text
import unittest
from unittest import mock
from kfp.deprecated.cli.diagnose_me import gcp
from kfp.deprecated.cli.diagnose_me import utility


class GoogleCloudTest(unittest.TestCase):

    @mock.patch.object(gcp, 'execute_gcloud_command', autospec=True)
    def test_project_configuration_gcloud(self, mock_execute_gcloud_command):
        """Tests gcloud commands."""
        gcp.get_gcp_configuration(gcp.Commands.GET_APIS)
        mock_execute_gcloud_command.assert_called_once_with(
            ['services', 'list'], project_id=None, human_readable=False)

    @mock.patch.object(gcp, 'execute_gsutil_command', autospec=True)
    def test_project_configuration_gsutil(self, mock_execute_gsutil_command):
        """Test Gsutil commands."""
        gcp.get_gcp_configuration(gcp.Commands.GET_STORAGE_BUCKETS)
        mock_execute_gsutil_command.assert_called_once_with(['ls'],
                                                            project_id=None)

    def test_Commands(self):
        """Verify commands are formaated properly."""
        for command in gcp.Commands:
            self.assertIsInstance(gcp._command_string[command], Text)
            self.assertNotIn('\t', gcp._command_string[command])
            self.assertNotIn('\n', gcp._command_string[command])

    @mock.patch.object(utility, 'ExecutorResponse', autospec=True)
    def test_execute_gsutil_command(self, mock_executor_response):
        """Test execute_gsutil_command."""
        gcp.execute_gsutil_command(
            [gcp._command_string[gcp.Commands.GET_STORAGE_BUCKETS]])
        mock_executor_response().execute_command.assert_called_once_with(
            ['gsutil', 'ls'])

        gcp.execute_gsutil_command(
            [gcp._command_string[gcp.Commands.GET_STORAGE_BUCKETS]],
            project_id='test_project')
        mock_executor_response().execute_command.assert_called_with(
            ['gsutil', 'ls', '-p', 'test_project'])

    @mock.patch.object(utility, 'ExecutorResponse', autospec=True)
    def test_execute_gcloud_command(self, mock_executor_response):
        """Test execute_gcloud_command."""
        gcp.execute_gcloud_command(
            gcp._command_string[gcp.Commands.GET_APIS].split(' '))
        mock_executor_response().execute_command.assert_called_once_with(
            ['gcloud', 'services', 'list', '--format', 'json'])

        gcp.execute_gcloud_command(
            gcp._command_string[gcp.Commands.GET_APIS].split(' '),
            project_id='test_project')
        # verify project id is added correctly
        mock_executor_response().execute_command.assert_called_with([
            'gcloud', 'services', 'list', '--format', 'json', '--project',
            'test_project'
        ])
        # verify human_readable removes json fromat flag
        gcp.execute_gcloud_command(
            gcp._command_string[gcp.Commands.GET_APIS].split(' '),
            project_id='test_project',
            human_readable=True)
        mock_executor_response().execute_command.assert_called_with(
            ['gcloud', 'services', 'list', '--project', 'test_project'])


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for KFP Registry RegistryClient."""

import json
import os
from unittest import mock

from absl.testing import parameterized
from kfp.registry import ApiAuth
from kfp.registry import RegistryClient
import requests

_DEFAULT_HOST = 'https://us-central1-kfp.pkg.dev/proj/repo'
_KFP_CONFIG_FILE = os.path.join(
    os.path.dirname(__file__), 'context/kfp_pkg_dev.json')


class RegistryClientTest(parameterized.TestCase):

    def setUp(self):
        super(RegistryClientTest, self).setUp()
        with open(_KFP_CONFIG_FILE, 'r') as f:
            read_value = f.read()
        self._mock_open = self.enter_context(
            mock.patch(
                'builtins.open',
                side_effect=[
                    mock.mock_open(read_data=read_value).return_value,
                    mock.mock_open(read_data='file_content').return_value
                ]))

    def test_is_ar_host_kfp(self):
        client = RegistryClient(
            host='https://us-central1-kfp.pkg.dev/proj/repo', auth=ApiAuth(''))
        self.assertEqual(client._is_ar_host(), True)

    def test_is_ar_host_false(self):
        client = RegistryClient(
            host='https://hub.docker.com/r/google/cloud-sdk', auth=ApiAuth(''))
        self.assertFalse(client._is_ar_host())

    def test_load_config(self):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        expected_config = {
            'host':
                host,
            'upload_url':
                host,
            'download_version_url':
                f'{host}/{{package_name}}/{{version}}',
            'download_tag_url':
                f'{host}/{{package_name}}/{{tag}}',
            'get_package_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}'),
            'list_packages_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages'),
            'delete_package_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}'),
            'get_tag_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/tags/{tag}'),
            'list_tags_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/tags'),
            'delete_tag_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/tags/{tag}'),
            'create_tag_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/tags?tagId={tag}'),
            'update_tag_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/tags/{tag}?updateMask=version'),
            'get_version_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/versions/{version}?view=FULL'),
            'list_versions_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/versions?view=FULL'),
            'delete_version_url':
                ('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/versions/{version}'),
            'package_format':
                ('projects/proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}'),
            'tag_format': ('projects/proj/locations/us-central1/repositories'
                           '/repo/packages/{package_name}/tags/{tag}'),
            'version_format':
                ('projects/proj/locations/us-central1/repositories'
                 '/repo/packages/{package_name}/versions/{version}')
        }
        self.assertEqual(self._mock_open.call_args_list[0][0],
                         (_KFP_CONFIG_FILE, 'r'))
        self.assertEqual(expected_config, client._config)

    @parameterized.parameters(
        {
            'version':
                'sha256:abcde12345',
            'tag':
                None,
            'file_name':
                None,
            'expected_url':
                'https://us-central1-kfp.pkg.dev/proj/repo/pack/sha256:abcde12345',
            'expected_file_name':
                'pack_abcde12345.yaml'
        },
        {
            'version':
                None,
            'tag':
                'tag1',
            'file_name':
                None,
            'expected_url':
                'https://us-central1-kfp.pkg.dev/proj/repo/pack/tag1',
            'expected_file_name':
                'pack_tag1.yaml'
        },
        {
            'version':
                None,
            'tag':
                'tag1',
            'file_name':
                'pipeline.yaml',
            'expected_url':
                'https://us-central1-kfp.pkg.dev/proj/repo/pack/tag1',
            'expected_file_name':
                'pipeline.yaml'
        },
    )
    @mock.patch('requests.get', autospec=True)
    def test_download_pipeline(self, mock_get, version, tag, file_name,
                               expected_url, expected_file_name):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.download_pipeline(
            package_name='pack', version=version, tag=tag, file_name=file_name)
        mock_get.assert_called_once_with(
            url=expected_url, data='', headers=None, auth=mock.ANY)
        self.assertEqual(self._mock_open.call_args_list[1][0],
                         (expected_file_name, 'wb'))

    def test_download_pipeline_version_error(self):
        client = RegistryClient(host=_DEFAULT_HOST, auth=ApiAuth(''))
        with self.assertRaises(ValueError):
            client.download_pipeline(package_name='pack', version='abcde12345')

    def test_download_pipeline_tag_error(self):
        client = RegistryClient(host=_DEFAULT_HOST, auth=ApiAuth(''))
        with self.assertRaises(ValueError):
            client.download_pipeline(
                package_name='pack', tag='sha256:abcde12345')

    @parameterized.parameters(
        {
            'tags': 'tag1',
            'expected_tags': 'tag1'
        },
        {
            'tags': ['tag1', 'tag2'],
            'expected_tags': 'tag1,tag2'
        },
    )
    @mock.patch('requests.post', autospec=True)
    def test_upload_pipeline(self, mock_post, tags, expected_tags):
        mock_post.return_value.text = 'package_name/sha256:abcde12345'
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        package_name, version = client.upload_pipeline(
            file_name='pipeline.yaml',
            tags=tags,
            extra_headers={'description': 'nothing'})
        mock_post.assert_called_once_with(
            url=host,
            data={'tags': expected_tags},
            headers={'description': 'nothing'},
            files={'content': mock.ANY},
            auth=mock.ANY)
        self.assertEqual(self._mock_open.call_args_list[1][0],
                         ('pipeline.yaml', 'rb'))
        self.assertEqual(package_name, 'package_name')
        self.assertEqual(version, 'sha256:abcde12345')

    @mock.patch('requests.get', autospec=True)
    def test_get_package(self, mock_get):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.get_package('pack')
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_list_packages(self, mock_get):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.list_packages()
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_list_packages_empty(self, mock_get):
        host = _DEFAULT_HOST
        mock_response = requests.Response()
        mock_response.status_code = 200
        mock_response._content = json.dumps({}).encode('utf-8')
        mock_get.return_value = mock_response
        client = RegistryClient(host=host, auth=ApiAuth(''))
        packages = client.list_packages()
        self.assertEqual(packages, {})
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.delete', autospec=True)
    def test_delete_package(self, mock_delete):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.delete_package('pack')
        mock_delete.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_get_version(self, mock_get):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.get_version('pack', 'sha256:abcde12345')
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/versions/sha256:abcde12345?view=FULL'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_list_versions(self, mock_get):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.list_versions('pack')
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/versions?view=FULL'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_list_versions_empty(self, mock_get):
        host = _DEFAULT_HOST
        mock_response = requests.Response()
        mock_response.status_code = 200
        mock_response._content = json.dumps({}).encode('utf-8')
        mock_get.return_value = mock_response
        client = RegistryClient(host=host, auth=ApiAuth(''))
        versions = client.list_versions('pack')
        self.assertEqual(versions, {})
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/versions?view=FULL'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.delete', autospec=True)
    def test_delete_version(self, mock_delete):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.delete_version('pack', 'sha256:abcde12345')
        mock_delete.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/versions/sha256:abcde12345'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_get_tag(self, mock_get):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.get_tag('pack', 'tag1')
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/tags/tag1'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_list_tags(self, mock_get):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.list_tags('pack')
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/tags'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.get', autospec=True)
    def test_list_tags_empty(self, mock_get):
        host = _DEFAULT_HOST
        mock_response = requests.Response()
        mock_response.status_code = 200
        mock_response._content = json.dumps({}).encode('utf-8')
        mock_get.return_value = mock_response
        client = RegistryClient(host=host, auth=ApiAuth(''))
        tags = client.list_tags('pack')
        self.assertEqual(tags, {})
        mock_get.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/tags'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.delete', autospec=True)
    def test_delete_tag(self, mock_delete):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.delete_tag('pack', 'tag1')
        mock_delete.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/tags/tag1'),
            data='',
            headers=None,
            auth=mock.ANY)

    @mock.patch('requests.post', autospec=True)
    def test_create_tag(self, mock_post):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.create_tag('pack', 'sha256:abcde12345', 'tag1')
        expected_data = json.dumps({
            'name':
                '',
            'version': ('projects/proj/locations/us-central1/repositories'
                        '/repo/packages/pack/versions/sha256:abcde12345')
        })
        mock_post.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/tags?tagId=tag1'),
            data=expected_data,
            headers={
                'Content-type': 'application/json',
            },
            auth=mock.ANY)

    @mock.patch('requests.patch', autospec=True)
    def test_update_tag(self, mock_patch):
        host = _DEFAULT_HOST
        client = RegistryClient(host=host, auth=ApiAuth(''))
        client.update_tag('pack', 'sha256:abcde12345', 'tag1')
        expected_data = json.dumps({
            'name':
                '',
            'version': ('projects/proj/locations/us-central1/repositories'
                        '/repo/packages/pack/versions/sha256:abcde12345')
        })
        mock_patch.assert_called_once_with(
            url=('https://artifactregistry.googleapis.com/v1/projects/'
                 'proj/locations/us-central1/repositories'
                 '/repo/packages/pack/tags/tag1?updateMask=version'),
            data=expected_data,
            headers={
                'Content-type': 'application/json',
            },
            auth=mock.ANY)



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional
import warnings

from ._container_op import BaseOp, ContainerOp
from . import _pipeline_param


class Resource(object):
    """A wrapper over Argo ResourceTemplate definition object
    (io.argoproj.workflow.v1alpha1.ResourceTemplate) which is used to represent
    the `resource` property in argo's workflow template
    (io.argoproj.workflow.v1alpha1.Template)."""
    swagger_types = {
        "action": "str",
        "merge_strategy": "str",
        "success_condition": "str",
        "failure_condition": "str",
        "set_owner_reference": "bool",
        "manifest": "str",
        "flags": "list[str]"
    }
    openapi_types = {
        "action": "str",
        "merge_strategy": "str",
        "success_condition": "str",
        "failure_condition": "str",
        "set_owner_reference": "bool",
        "manifest": "str",
        "flags": "list[str]"
    }
    attribute_map = {
        "action": "action",
        "merge_strategy": "mergeStrategy",
        "success_condition": "successCondition",
        "failure_condition": "failureCondition",
        "set_owner_reference": "setOwnerReference",
        "manifest": "manifest",
        "flags": "flags"
    }

    def __init__(self,
                 action: str = None,
                 merge_strategy: str = None,
                 success_condition: str = None,
                 failure_condition: str = None,
                 set_owner_reference: bool = None,
                 manifest: str = None,
                 flags: Optional[List[str]] = None):
        """Create a new instance of Resource."""
        self.action = action
        self.merge_strategy = merge_strategy
        self.success_condition = success_condition
        self.failure_condition = failure_condition
        self.set_owner_reference = set_owner_reference
        self.manifest = manifest
        self.flags = flags


class ResourceOp(BaseOp):
    """Represents an op which will be translated into a resource template.

    TODO(https://github.com/kubeflow/pipelines/issues/4822): Determine the
        stability level of this feature.

    Args:
      k8s_resource: A k8s resource which will be submitted to the cluster
      action: One of "create"/"delete"/"apply"/"patch" (default is "create")
      merge_strategy: The merge strategy for the "apply" action
      success_condition: The successCondition of the template
      failure_condition: The failureCondition of the template
          For more info see:
          https://github.com/argoproj/argo-workflows/blob/master/examples/k8s-jobs.yaml
      attribute_outputs: Maps output labels to resource's json paths,
        similarly to file_outputs of ContainerOp
      kwargs: name, sidecars. See BaseOp definition

    Raises:
      ValueError: if not inside a pipeline
          if the name is an invalid string
          if no k8s_resource is provided
          if merge_strategy is set without "apply" action
    """

    def __init__(self,
                 k8s_resource=None,
                 action: str = "create",
                 merge_strategy: str = None,
                 success_condition: str = None,
                 failure_condition: str = None,
                 set_owner_reference: bool = None,
                 attribute_outputs: Optional[Dict[str, str]] = None,
                 flags: Optional[List[str]] = None,
                 **kwargs):

        super().__init__(**kwargs)
        self.attrs_with_pipelineparams = list(self.attrs_with_pipelineparams)
        self.attrs_with_pipelineparams.extend(
            ["_resource", "k8s_resource", "attribute_outputs"])

        if k8s_resource is None:
            raise ValueError("You need to provide a k8s_resource.")

        if action == "delete":
            warnings.warn(
                'Please use `kubernetes_resource_delete_op` instead of '
                '`ResourceOp(action="delete")`', DeprecationWarning)

        if merge_strategy and action != "apply":
            raise ValueError(
                "You can't set merge_strategy when action != 'apply'")

        # if action is delete, there should not be any outputs, success_condition,
        # and failure_condition
        if action == "delete" and (success_condition or failure_condition or
                                   attribute_outputs):
            raise ValueError(
                "You can't set success_condition, failure_condition, or "
                "attribute_outputs when action == 'delete'")

        if action == "delete" and flags is None:
            flags = ["--wait=false"]
        init_resource = {
            "action": action,
            "merge_strategy": merge_strategy,
            "success_condition": success_condition,
            "failure_condition": failure_condition,
            "set_owner_reference": set_owner_reference,
            "flags": flags
        }
        # `resource` prop in `io.argoproj.workflow.v1alpha1.Template`
        self._resource = Resource(**init_resource)

        self.k8s_resource = k8s_resource

        # if action is delete, there should not be any outputs, success_condition,
        # and failure_condition
        if action == "delete":
            self.attribute_outputs = {}
            self.outputs = {}
            self.output = None
            return

        # Set attribute_outputs
        extra_attribute_outputs = \
            attribute_outputs if attribute_outputs else {}
        self.attribute_outputs = \
            self.attribute_outputs if hasattr(self, "attribute_outputs") \
            else {}
        self.attribute_outputs.update(extra_attribute_outputs)
        # Add name and manifest if not specified by the user
        if "name" not in self.attribute_outputs:
            self.attribute_outputs["name"] = "{.metadata.name}"
        if "manifest" not in self.attribute_outputs:
            self.attribute_outputs["manifest"] = "{}"

        # Set outputs
        self.outputs = {
            name: _pipeline_param.PipelineParam(name, op_name=self.name)
            for name in self.attribute_outputs.keys()
        }
        # If user set a single attribute_output, set self.output as that
        # parameter, else set it as the resource name
        self.output = self.outputs["name"]
        if len(extra_attribute_outputs) == 1:
            self.output = self.outputs[list(extra_attribute_outputs)[0]]

    @property
    def resource(self):
        """`Resource` object that represents the `resource` property in
        `io.argoproj.workflow.v1alpha1.Template`."""
        return self._resource

    def delete(self, flags: Optional[List[str]] = None):
        """Returns a ResourceOp which deletes the resource."""
        if self.resource.action == "delete":
            raise ValueError("This operation is already a resource deletion.")

        if isinstance(self.k8s_resource, dict):
            kind = self.k8s_resource["kind"]
        else:
            kind = self.k8s_resource.kind

        return kubernetes_resource_delete_op(
            name=self.outputs["name"],
            kind=kind,
            flags=flags or ["--wait=false"])


def kubernetes_resource_delete_op(
    name: str,
    kind: str,
    namespace: str = None,
    flags: Optional[List[str]] = None,
) -> ContainerOp:
    """Operation that deletes a Kubernetes resource.

    Outputs:
      name: The name of the deleted resource
    """

    command = [
        "kubectl", "delete",
        str(kind),
        str(name), "--ignore-not-found", "--output", "name"
    ]
    if namespace:
        command.extend(["--namespace", str(namespace)])
    if flags:
        command.extend(flags)

    result = ContainerOp(
        name="kubernetes_resource_delete",
        image="gcr.io/cloud-builders/kubectl",
        command=command,
    )
    return result



# Copyright 2021-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.base_component."""

import unittest
from unittest.mock import patch

from kfp import dsl
from kfp.components import pipeline_task
from kfp.components import placeholders
from kfp.components import PythonComponent
from kfp.components import structures

component_op = PythonComponent(
    # dummy python_func not used in behavior that is being tested
    python_func=lambda: None,
    component_spec=structures.ComponentSpec(
        name='component_1',
        implementation=structures.Implementation(
            container=structures.ContainerSpecImplementation(
                image='alpine',
                command=[
                    'sh',
                    '-c',
                    'set -ex\necho "$0" "$1" "$2" > "$3"',
                    placeholders.InputValuePlaceholder(input_name='input1'),
                    placeholders.InputValuePlaceholder(input_name='input2'),
                    placeholders.InputValuePlaceholder(input_name='input3'),
                    placeholders.OutputPathPlaceholder(output_name='output1'),
                ],
            )),
        inputs={
            'input1':
                structures.InputSpec(type='String'),
            'input2':
                structures.InputSpec(type='Integer'),
            'input3':
                structures.InputSpec(type='Float', default=3.14, optional=True),
            'input4':
                structures.InputSpec(
                    type='Optional[Float]', default=None, optional=True),
        },
        outputs={
            'output1': structures.OutputSpec(type='String'),
        },
    ))


class BaseComponentTest(unittest.TestCase):

    @patch.object(pipeline_task, 'PipelineTask', autospec=True)
    def test_instantiate_component_with_keyword_arguments(
            self, mock_PipelineTask):

        component_op(input1='hello', input2=100, input3=1.23, input4=3.21)

        mock_PipelineTask.assert_called_once_with(
            component_spec=component_op.component_spec,
            args={
                'input1': 'hello',
                'input2': 100,
                'input3': 1.23,
                'input4': 3.21,
            })

    @patch.object(pipeline_task, 'PipelineTask', autospec=True)
    def test_instantiate_component_omitting_arguments_with_default(
            self, mock_PipelineTask):

        component_op(input1='hello', input2=100)

        mock_PipelineTask.assert_called_once_with(
            component_spec=component_op.component_spec,
            args={
                'input1': 'hello',
                'input2': 100,
            })

    def test_instantiate_component_with_positional_arugment(self):
        with self.assertRaisesRegex(
                TypeError,
                'Components must be instantiated using keyword arguments.'
                r' Positional parameters are not allowed \(found 3 such'
                r' parameters for component "component-1"\).'):
            component_op('abc', 1, 2.3)

    def test_instantiate_component_with_unexpected_keyword_arugment(self):
        with self.assertRaisesRegex(
                TypeError,
                r'component-1\(\) got an unexpected keyword argument "input0".'
        ):
            component_op(input1='abc', input2=1, input3=2.3, input0='extra')

    def test_instantiate_component_with_missing_arugments(self):
        with self.assertRaisesRegex(
                TypeError,
                r'component-1\(\) missing 1 required argument: input1.'):
            component_op(input2=1)

        with self.assertRaisesRegex(
                TypeError,
                r'component-1\(\) missing 2 required arguments: input1, input2.'
        ):
            component_op()


class BlockPipelineTaskRegistration(unittest.TestCase):

    def test_mutating_decorator(self):

        def call_pipeline_spec(component):
            component.pipeline_spec
            return component

        @dsl.component
        def identity(text: str) -> str:
            return text

        @dsl.pipeline
        def pipeline_custom_job():
            modified_identity = call_pipeline_spec(identity)
            modified_identity(text='text')

        self.assertEqual(len(pipeline_custom_job.pipeline_spec.components), 1)
        self.assertEqual(
            len(pipeline_custom_job.pipeline_spec.deployment_spec), 1)

    def test_call_directly_in_pipeline(self):

        @dsl.component
        def identity(text: str) -> str:
            return text

        @dsl.pipeline
        def pipeline_custom_job():
            identity.pipeline_spec
            identity(text='text')

        self.assertEqual(len(pipeline_custom_job.pipeline_spec.components), 1)
        self.assertEqual(
            len(pipeline_custom_job.pipeline_spec.deployment_spec), 1)


if __name__ == '__main__':
    unittest.main()



"""The `kfp.client` module contains the KFP API client."""
# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'Client',
]

from kfp.client.client import Client
from kfp.client.set_volume_credentials import \
    ServiceAccountTokenVolumeCredentials
from kfp.client.token_credentials_base import TokenCredentialsBase

KF_PIPELINES_SA_TOKEN_ENV = 'KF_PIPELINES_SA_TOKEN_PATH'
KF_PIPELINES_SA_TOKEN_PATH = '/var/run/secrets/kubeflow/pipelines/token'



# Copyright 2019-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import click
from kfp import client
from kfp.cli import output
from kfp.cli.utils import deprecated_alias_group
from kfp.cli.utils import parsing


@click.group(
    cls=deprecated_alias_group.deprecated_alias_group_factory({
        'upload': 'create',
        'upload-version': 'create-version'
    }))
def pipeline():
    """Manage pipeline resources."""


@pipeline.command()
@click.option(
    '-p',
    '--pipeline-name',
    help=parsing.get_param_descr(client.Client.upload_pipeline,
                                 'pipeline_name'))
@click.option(
    '-d',
    '--description',
    help=parsing.get_param_descr(client.Client.upload_pipeline, 'description'))
@click.argument('package-file')
@click.pass_context
def create(ctx: click.Context,
           pipeline_name: str,
           package_file: str,
           description: str = None):
    """Upload a pipeline."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    pipeline = client_obj.upload_pipeline(package_file, pipeline_name,
                                          description)
    output.print_output(
        pipeline,
        output.ModelType.PIPELINE,
        output_format,
    )


either_option_required = 'Either --pipeline-id or --pipeline-name is required.'


@pipeline.command()
@click.argument('package-file', type=click.Path(exists=True, dir_okay=False))
@click.option(
    '-v',
    '--pipeline-version',
    help=parsing.get_param_descr(client.Client.upload_pipeline_version,
                                 'pipeline_version_name'),
    required=True,
)
@click.option(
    '-p',
    '--pipeline-id',
    required=False,
    help=parsing.get_param_descr(client.Client.upload_pipeline_version,
                                 'pipeline_id') + ' ' + either_option_required)
@click.option(
    '-n',
    '--pipeline-name',
    required=False,
    help=parsing.get_param_descr(client.Client.upload_pipeline_version,
                                 'pipeline_name') + ' ' + either_option_required
)
@click.option(
    '-d',
    '--description',
    help=parsing.get_param_descr(client.Client.upload_pipeline_version,
                                 'description'))
@click.pass_context
def create_version(ctx: click.Context,
                   package_file: str,
                   pipeline_version: str,
                   pipeline_id: Optional[str] = None,
                   pipeline_name: Optional[str] = None,
                   description: Optional[str] = None):
    """Upload a version of a pipeline."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']
    if bool(pipeline_id) == bool(pipeline_name):
        raise ValueError(either_option_required)
    if pipeline_name is not None:
        pipeline_id = client_obj.get_pipeline_id(name=pipeline_name)
        if pipeline_id is None:
            raise ValueError(
                f"Can't find a pipeline with name: {pipeline_name}")
    version = client_obj.upload_pipeline_version(
        pipeline_package_path=package_file,
        pipeline_version_name=pipeline_version,
        pipeline_id=pipeline_id,
        pipeline_name=pipeline_name,
        description=description)
    output.print_output(
        version,
        output.ModelType.PIPELINE,
        output_format,
    )


@pipeline.command()
@click.option(
    '--page-token',
    default='',
    help=parsing.get_param_descr(client.Client.list_pipelines, 'page_token'))
@click.option(
    '-m',
    '--max-size',
    default=100,
    help=parsing.get_param_descr(client.Client.list_pipelines, 'page_size'))
@click.option(
    '--sort-by',
    default='created_at desc',
    help=parsing.get_param_descr(client.Client.list_pipelines, 'sort_by'))
@click.option(
    '--filter',
    help=parsing.get_param_descr(client.Client.list_pipelines, 'filter'))
@click.pass_context
def list(ctx: click.Context, page_token: str, max_size: int, sort_by: str,
         filter: str):
    """List pipelines."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    response = client_obj.list_pipelines(
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    output.print_output(
        response.pipelines or [],
        output.ModelType.PIPELINE,
        output_format,
    )


@pipeline.command()
@click.argument('pipeline-id')
@click.option(
    '--page-token',
    default='',
    help=parsing.get_param_descr(client.Client.list_pipeline_versions,
                                 'page_token'))
@click.option(
    '-m',
    '--max-size',
    default=100,
    help=parsing.get_param_descr(client.Client.list_pipeline_versions,
                                 'page_size'))
@click.option(
    '--sort-by',
    default='created_at desc',
    help=parsing.get_param_descr(client.Client.list_pipeline_versions,
                                 'sort_by'))
@click.option(
    '--filter',
    help=parsing.get_param_descr(client.Client.list_pipeline_versions,
                                 'filter'))
@click.pass_context
def list_versions(ctx: click.Context, pipeline_id: str, page_token: str,
                  max_size: int, sort_by: str, filter: str):
    """List versions of a pipeline."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    response = client_obj.list_pipeline_versions(
        pipeline_id,
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    output.print_output(
        response.pipeline_versions or [],
        output.ModelType.PIPELINE_VERSION,
        output_format,
    )


@pipeline.command()
@click.argument('pipeline-id')
@click.argument('version-id')
@click.pass_context
def delete_version(ctx: click.Context, pipeline_id: str, version_id: str):
    """Delete a version of a pipeline."""
    confirmation = f'Are you sure you want to delete pipeline version {version_id}?'
    if not click.confirm(confirmation):
        return

    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    client_obj.delete_pipeline_version(
        pipeline_id=pipeline_id, pipeline_version_id=version_id)
    output.print_deleted_text('pipeline version', version_id, output_format)


@pipeline.command()
@click.argument('pipeline-id')
@click.pass_context
def get(ctx: click.Context, pipeline_id: str):
    """Get information about a pipeline."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    pipeline = client_obj.get_pipeline(pipeline_id)
    output.print_output(
        pipeline,
        output.ModelType.PIPELINE,
        output_format,
    )


@pipeline.command()
@click.argument('pipeline-id')
@click.argument('version-id')
@click.pass_context
def get_version(ctx: click.Context, pipeline_id: str, version_id: str):
    """Get information about a version of a pipeline."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    version = client_obj.get_pipeline_version(
        pipeline_id=pipeline_id, pipeline_version_id=version_id)
    output.print_output(
        version,
        output.ModelType.PIPELINE,
        output_format,
    )


@pipeline.command()
@click.argument('pipeline-id')
@click.pass_context
def delete(ctx: click.Context, pipeline_id: str):
    """Delete a pipeline."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    confirmation = f'Are you sure you want to delete pipeline {pipeline_id}?'
    if not click.confirm(confirmation):
        return

    client_obj.delete_pipeline(pipeline_id)
    output.print_deleted_text('pipeline', pipeline_id, output_format)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python function-based component."""

from typing import Callable

from kfp import components
from kfp.components import structures


class PythonComponent(components.BaseComponent):
    """A component defined via Python function.

    **Note:** ``PythonComponent`` is not intended to be used to construct components directly. Use ``@kfp.dsl.component`` instead.

    Args:
        component_spec: Component definition.
        python_func: Python function that becomes the implementation of this component.
    """

    def __init__(
        self,
        component_spec: structures.ComponentSpec,
        python_func: Callable,
    ):
        super().__init__(component_spec=component_spec)
        self.python_func = python_func

        self._prevent_using_output_lists_of_artifacts()

    def execute(self, **kwargs):
        """Executes the Python function that defines the component."""
        return self.python_func(**kwargs)



# Copyright 2018-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import abc
from collections import OrderedDict
import inspect
from typing import (Any, cast, Dict, get_type_hints, List, Mapping,
                    MutableMapping, MutableSequence, Sequence, Type, TypeVar,
                    Union)

T = TypeVar('T')


def verify_object_against_type(x: Any, typ: Type[T]) -> T:
    """Verifies that the object is compatible to the specified type (types from
    the typing package can be used)."""
    #TODO: Merge with parse_object_from_struct_based_on_type which has almost the same code
    if typ is type(None):
        if x is None:
            return x
        else:
            raise TypeError(f'Error: Object "{x}" is not None.')

    if typ is Any or type(typ) is TypeVar:
        return x

    try:  #isinstance can fail for generics
        if isinstance(x, typ):
            return cast(typ, x)
    except Exception:
        pass

    if hasattr(typ, '__origin__'):  #Handling generic types
        if typ.__origin__ is Union:  #Optional == Union
            exception_map = {}
            possible_types = typ.__args__
            if type(
                    None
            ) in possible_types and x is None:  #Shortcut for Optional[] tests. Can be removed, but the exceptions will be more noisy.
                return x
            for possible_type in possible_types:
                try:
                    verify_object_against_type(x, possible_type)
                    return x
                except Exception as ex:
                    exception_map[possible_type] = ex
            #exception_lines = ['Exception for type {}: {}.'.format(t, e) for t, e in exception_map.items()]
            exception_lines = [str(e) for t, e in exception_map.items()]
            exception_lines.append(
                f'Error: Object "{x}" is incompatible with type "{typ}".')
            raise TypeError('\n'.join(exception_lines))

        #not Union => not None
        if x is None:
            raise TypeError(
                f'Error: None object is incompatible with type {typ}')

        generic_type = typ.__origin__
        if generic_type in [
                list, List, abc.Sequence, abc.MutableSequence, Sequence,
                MutableSequence
        ] and type(x) is not str:  #! str is also Sequence
            if not isinstance(x, generic_type):
                raise TypeError(
                    f'Error: Object "{x}" is incompatible with type "{typ}"')

            # In Python 3.9 typ.__args__ does not exist when the generic type does not have subscripts
            type_args = typ.__args__ if getattr(
                typ, '__args__', None) is not None else (Any, Any)
            inner_type = type_args[0]
            for item in x:
                verify_object_against_type(item, inner_type)
            return x

        elif generic_type in [
                dict, Dict, abc.Mapping, abc.MutableMapping, Mapping,
                MutableMapping, OrderedDict
        ]:
            if not isinstance(x, generic_type):
                raise TypeError(
                    f'Error: Object "{x}" is incompatible with type "{typ}"')

            # In Python 3.9 typ.__args__ does not exist when the generic type does not have subscripts
            type_args = typ.__args__ if getattr(
                typ, '__args__', None) is not None else (Any, Any)
            inner_key_type = type_args[0]
            inner_value_type = type_args[1]
            for k, v in x.items():
                verify_object_against_type(k, inner_key_type)
                verify_object_against_type(v, inner_value_type)
            return x

        else:
            raise TypeError(
                f'Error: Unsupported generic type "{typ}". type.__origin__ or type.__extra__ == "{generic_type}"'
            )

    raise TypeError(f'Error: Object "{x}" is incompatible with type "{typ}"')


def parse_object_from_struct_based_on_type(struct: Any, typ: Type[T]) -> T:
    """Constructs an object from structure (usually dict) based on type.

    Supports list and dict types from the typing package plus Optional[]
    and Union[] types. If some type is a class that has .from_dict class
    method, that method is used for object construction.
    """
    if typ is type(None):
        if struct is None:
            return None
        else:
            raise TypeError(f'Error: Structure "{struct}" is not None.')

    if typ is Any or type(typ) is TypeVar:
        return struct

    try:  #isinstance can fail for generics
        #if (isinstance(struct, typ)
        #    and not (typ is Sequence and type(struct) is str) #! str is also Sequence
        #    and not (typ is int and type(struct) is bool) #! bool is int
        #):
        if type(struct) is typ:
            return struct
    except:
        pass
    if hasattr(typ, 'from_dict'):
        try:  #More informative errors
            return typ.from_dict(struct)
        except Exception as ex:
            raise TypeError(
                f'Error: {typ.__name__}.from_dict(struct={struct}) failed with exception:\n{str(ex)}'
            )
    if hasattr(typ, '__origin__'):  #Handling generic types
        if typ.__origin__ is Union:  #Optional == Union
            results = {}
            exception_map = {}
            # In Python 3.9 typ.__args__ does not exist when the generic type does not have subscripts
            # Union without subscripts seems useless, but semantically it should be the same as Any.
            possible_types = list(getattr(typ, '__args__', [Any]))
            #if type(None) in possible_types and struct is None: #Shortcut for Optional[] tests. Can be removed, but the exceptions will be more noisy.
            #    return None

            for possible_type in possible_types:
                try:
                    obj = parse_object_from_struct_based_on_type(
                        struct, possible_type)
                    results[possible_type] = obj
                except Exception as ex:
                    if isinstance(ex, TypeError):
                        exception_map[possible_type] = ex
                    else:
                        exception_map[
                            possible_type] = f'Unexpected exception when trying to convert structure "{struct}" to type "{typ}": {type(ex)}: {ex}'

            #Single successful parsing.
            if len(results) == 1:
                return list(results.values())[0]

            if len(results) > 1:
                raise TypeError(
                    f'Error: Structure "{struct}" is ambiguous. It can be parsed to multiple types: {list(results.keys())}.'
                )

            exception_lines = [str(e) for t, e in exception_map.items()]
            exception_lines.append(
                f'Error: Structure "{struct}" is incompatible with type "{typ}" - none of the types in Union are compatible.'
            )
            raise TypeError('\n'.join(exception_lines))
        #not Union => not None
        if struct is None:
            raise TypeError(
                f'Error: None structure is incompatible with type {typ}')

        generic_type = typ.__origin__
        if generic_type in [
                list, List, abc.Sequence, abc.MutableSequence, Sequence,
                MutableSequence
        ] and type(struct) is not str:  #! str is also Sequence
            if not isinstance(struct, generic_type):
                raise TypeError(
                    f'Error: Structure "{struct}" is incompatible with type "{typ}" - it does not have list type.'
                )

            # In Python 3.9 typ.__args__ does not exist when the generic type does not have subscripts
            type_args = typ.__args__ if getattr(
                typ, '__args__', None) is not None else (Any, Any)
            inner_type = type_args[0]
            return [
                parse_object_from_struct_based_on_type(item, inner_type)
                for item in struct
            ]

        elif generic_type in [
                dict, Dict, abc.Mapping, abc.MutableMapping, Mapping,
                MutableMapping, OrderedDict
        ]:
            if not isinstance(struct, generic_type):
                raise TypeError(
                    f'Error: Structure "{struct}" is incompatible with type "{typ}" - it does not have dict type.'
                )

            # In Python 3.9 typ.__args__ does not exist when the generic type does not have subscripts
            type_args = typ.__args__ if getattr(
                typ, '__args__', None) is not None else (Any, Any)
            inner_key_type = type_args[0]
            inner_value_type = type_args[1]
            return {
                parse_object_from_struct_based_on_type(k, inner_key_type):
                parse_object_from_struct_based_on_type(v, inner_value_type)
                for k, v in struct.items()
            }

        else:
            raise TypeError(
                f'Error: Unsupported generic type "{typ}". type.__origin__ or type.__extra__ == "{generic_type}"'
            )

    raise TypeError(
        f'Error: Structure "{struct}" is incompatible with type "{typ}". Structure is not the instance of the type, the type does not have .from_dict method and is not generic.'
    )


def convert_object_to_struct(obj, serialized_names: Mapping[str, str] = {}):
    """Converts an object to structure (usually a dict).

    Serializes all properties that do not start with underscores. If the
    type of some property is a class that has .to_dict class method,
    that method is used for conversion. Used by the ModelBase class.
    """
    signature = inspect.signature(obj.__init__)  #Needed for default values
    result = {}
    for python_name in signature.parameters:  #TODO: Make it possible to specify the field ordering regardless of the presence of default values
        value = getattr(obj, python_name)
        if python_name.startswith('_'):
            continue
        attr_name = serialized_names.get(python_name, python_name)
        if hasattr(value, 'to_dict'):
            result[attr_name] = value.to_dict()
        elif isinstance(value, list):
            result[attr_name] = [
                (x.to_dict() if hasattr(x, 'to_dict') else x) for x in value
            ]
        elif isinstance(value, dict):
            result[attr_name] = {
                k: (v.to_dict() if hasattr(v, 'to_dict') else v)
                for k, v in value.items()
            }
        else:
            param = signature.parameters.get(python_name, None)
            if param is None or param.default == inspect.Parameter.empty or value != param.default:
                result[attr_name] = value

    return result


def parse_object_from_struct_based_on_class_init(
        cls: Type[T],
        struct: Mapping,
        serialized_names: Mapping[str, str] = {}) -> T:
    """Constructs an object of specified class from structure (usually dict)
    using the class.__init__ method. Converts all constructor arguments to
    appropriate types based on the __init__ type hints. Used by the ModelBase
    class.

    Arguments:

    serialized_names: specifies the mapping between __init__ parameter names and the structure key names for cases where these names are different (due to language syntax clashes or style differences).
    """
    parameter_types = get_type_hints(
        cls.__init__)  #Properlty resolves forward references

    serialized_names_to_pythonic = {v: k for k, v in serialized_names.items()}
    #If a pythonic name has a different original name, we forbid the pythonic name in the structure. Otherwise, this function would accept "python-styled" structures that should be invalid
    forbidden_struct_keys = set(
        serialized_names_to_pythonic.values()).difference(
            serialized_names_to_pythonic.keys())
    args = {}
    for original_name, value in struct.items():
        if original_name in forbidden_struct_keys:
            raise ValueError(
                f'Use "{serialized_names[original_name]}" key instead of pythonic key "{original_name}" in the structure: {struct}.'
            )
        python_name = serialized_names_to_pythonic.get(original_name,
                                                       original_name)
        param_type = parameter_types.get(python_name, None)
        if param_type is not None:
            args[python_name] = parse_object_from_struct_based_on_type(
                value, param_type)
        else:
            args[python_name] = value

    return cls(**args)


class ModelBase:
    """Base class for types that can be converted to JSON-like dict structures
    or constructed from such structures. The object fields, their types and
    default values are taken from the __init__ method arguments. Override the
    _serialized_names mapping to control the key names of the serialized
    structures.

    The derived class objects will have the .from_dict and .to_dict methods for conversion to or from structure. The base class constructor accepts the arguments map, checks the argument types and sets the object field values.

    Example derived class:

    class TaskSpec(ModelBase):
        _serialized_names = {
            'component_ref': 'componentRef',
            'is_enabled': 'isEnabled',
        }

        def __init__(self,
            component_ref: ComponentReference,
            arguments: Optional[Mapping[str, ArgumentType]] = None,
            is_enabled: Optional[Union[ArgumentType, EqualsPredicate, NotEqualsPredicate]] = None, #Optional property with default value
        ):
            super().__init__(locals()) #Calling the ModelBase constructor to check the argument types and set the object field values.

    task_spec = TaskSpec.from_dict("{'componentRef': {...}, 'isEnabled: {'and': {...}}}") # = instance of TaskSpec
    task_struct = task_spec.to_dict() #= "{'componentRef': {...}, 'isEnabled: {'and': {...}}}"
    """
    _serialized_names = {}

    def __init__(self, args):
        parameter_types = get_type_hints(self.__class__.__init__)
        field_values = {
            k: v
            for k, v in args.items()
            if k != 'self' and not k.startswith('_')
        }
        for k, v in field_values.items():
            parameter_type = parameter_types.get(k, None)
            if parameter_type is not None:
                try:
                    verify_object_against_type(v, parameter_type)
                except Exception as e:
                    raise TypeError(
                        f'Argument for {k} is not compatible with type "{parameter_type}". Exception: {e}'
                    )
        self.__dict__.update(field_values)

    @classmethod
    def from_dict(cls: Type[T], struct: Mapping) -> T:
        return parse_object_from_struct_based_on_class_init(
            cls, struct, serialized_names=cls._serialized_names)

    def to_dict(self) -> Dict[str, Any]:
        return convert_object_to_struct(
            self, serialized_names=self._serialized_names)

    def _get_field_names(self):
        return list(inspect.signature(self.__init__).parameters)

    def __repr__(self):
        return self.__class__.__name__ + '(' + ', '.join(
            param + '=' + repr(getattr(self, param))
            for param in self._get_field_names()) + ')'

    def __eq__(self, other):
        return self.__class__ == other.__class__ and {
            k: getattr(self, k) for k in self._get_field_names()
        } == {k: getattr(other, k) for k in other._get_field_names()}

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(repr(self))



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains data structures and functions for handling input and output
placeholders."""

import abc
import json
from typing import Any, Dict, List, Optional, Union

from kfp.components import utils
from kfp.components.types import type_utils


class Placeholder(abc.ABC):

    @abc.abstractmethod
    def _to_string(self) -> str:
        raise NotImplementedError

    def __str__(self) -> str:
        """Enables use of placeholders in f-strings.

        To be overridden by container placeholders ConcatPlaceholder and
        IfPresentPlaceholder, which cannot be used in an f-string.
        """
        return self._to_string()

    def __eq__(self, other: Any) -> bool:
        """Used for comparing placeholders in tests."""
        return isinstance(other,
                          self.__class__) and self.__dict__ == other.__dict__


class ExecutorInputPlaceholder(Placeholder):

    def _to_string(self) -> str:
        return '{{$}}'


class InputValuePlaceholder(Placeholder):

    def __init__(self, input_name: str) -> None:
        self.input_name = input_name

    def _to_string(self) -> str:
        return f"{{{{$.inputs.parameters['{self.input_name}']}}}}"


class InputListOfArtifactsPlaceholder(Placeholder):

    def __init__(self, input_name: str) -> None:
        self.input_name = input_name

    def _to_string(self) -> str:
        return f"{{{{$.inputs.artifacts['{self.input_name}']}}}}"

    def __getattribute__(self, name: str) -> Any:
        if name in {'name', 'uri', 'metadata', 'path'}:
            raise AttributeError(
                f'Cannot access an attribute on a list of artifacts in a Custom Container Component. Found reference to attribute {name!r} on {self.input_name!r}. Please pass the whole list of artifacts only.'
            )
        else:
            return object.__getattribute__(self, name)

    def __getitem__(self, k: int) -> None:
        raise KeyError(
            f'Cannot access individual artifacts in a list of artifacts. Found access to element {k} on {self.input_name!r}. Please pass the whole list of artifacts only.'
        )


class OutputListOfArtifactsPlaceholder(Placeholder):

    def __init__(self, input_name: str) -> None:
        self.output_name = input_name

    def _to_string(self) -> str:
        return f"{{{{$.outputs.artifacts['{self.output_name}']}}}}"

    def __getattribute__(self, name: str) -> Any:
        if name in {'name', 'uri', 'metadata', 'path'}:
            raise AttributeError(
                f'Cannot access an attribute on a list of artifacts in a Custom Container Component. Found reference to attribute {name!r} on {self.output_name!r}. Please pass the whole list of artifacts only.'
            )
        else:
            return object.__getattribute__(self, name)

    def __getitem__(self, k: int) -> None:
        raise KeyError(
            f'Cannot access individual artifacts in a list of artifacts. Found access to element {k} on {self.output_name!r}. Please pass the whole list of artifacts only.'
        )


class InputPathPlaceholder(Placeholder):

    def __init__(self, input_name: str) -> None:
        self.input_name = input_name

    def _to_string(self) -> str:
        return f"{{{{$.inputs.artifacts['{self.input_name}'].path}}}}"


class InputUriPlaceholder(Placeholder):

    def __init__(self, input_name: str) -> None:
        self.input_name = input_name

    def _to_string(self) -> str:
        return f"{{{{$.inputs.artifacts['{self.input_name}'].uri}}}}"


class InputMetadataPlaceholder(Placeholder):

    def __init__(self, input_name: str) -> None:
        self.input_name = input_name

    def _to_string(self) -> str:
        return f"{{{{$.inputs.artifacts['{self.input_name}'].metadata}}}}"

    def __getitem__(self, key: str) -> str:
        return f"{{{{$.inputs.artifacts['{self.input_name}'].metadata['{key}']}}}}"


class OutputParameterPlaceholder(Placeholder):

    def __init__(self, output_name: str) -> None:
        self.output_name = output_name

    def _to_string(self) -> str:
        return f"{{{{$.outputs.parameters['{self.output_name}'].output_file}}}}"


class OutputPathPlaceholder(Placeholder):

    def __init__(self, output_name: str) -> None:
        self.output_name = output_name

    def _to_string(self) -> str:
        return f"{{{{$.outputs.artifacts['{self.output_name}'].path}}}}"


class OutputUriPlaceholder(Placeholder):

    def __init__(self, output_name: str) -> None:
        self.output_name = output_name

    def _to_string(self) -> str:
        return f"{{{{$.outputs.artifacts['{self.output_name}'].uri}}}}"


class OutputMetadataPlaceholder(Placeholder):

    def __init__(self, output_name: str) -> None:
        self.output_name = output_name

    def _to_string(self) -> str:
        return f"{{{{$.outputs.artifacts['{self.output_name}'].metadata}}}}"

    def __getitem__(self, key: str) -> str:
        return f"{{{{$.outputs.artifacts['{self.output_name}'].metadata['{key}']}}}}"


class ConcatPlaceholder(Placeholder):
    """Placeholder for concatenating multiple strings. May contain other
    placeholders.

    Args:
        items: Elements to concatenate.

    Examples:
      ::

        @container_component
        def container_with_concat_placeholder(text1: str, text2: Output[Dataset],
                                              output_path: OutputPath(str)):
            return ContainerSpec(
                image='python:3.7',
                command=[
                    'my_program',
                    ConcatPlaceholder(['prefix-', text1, text2.uri])
                ],
                args=['--output_path', output_path]
            )
    """

    def __init__(self, items: List['CommandLineElement']) -> None:
        for item in items:
            if isinstance(item, IfPresentPlaceholder):
                item._validate_then_and_else_are_only_single_element()
        self.items = items

    def _to_dict(self) -> Dict[str, Any]:
        return {
            'Concat': [
                convert_command_line_element_to_string_or_struct(item)
                for item in self.items
            ]
        }

    def _to_string(self) -> str:
        return json.dumps(self._to_dict())

    def __str__(self) -> str:
        raise ValueError(
            f'Cannot use {self.__class__.__name__} in an f-string.')


class IfPresentPlaceholder(Placeholder):
    """Placeholder for handling cases where an input may or may not be passed.
    May contain other placeholders.

    Args:
        input_name: Name of the input/output.
        then: If the input/output specified in name is present, the command-line argument will be replaced at run-time by the value of then.
        else_: If the input/output specified in name is not present, the command-line argument will be replaced at run-time by the value of else_.

    Examples:
      ::

        @container_component
        def container_with_if_placeholder(output_path: OutputPath(str),
                                          dataset: Output[Dataset],
                                          optional_input: str = 'default'):
            return ContainerSpec(
                    image='python:3.7',
                    command=[
                        'my_program',
                        IfPresentPlaceholder(
                            input_name='optional_input',
                            then=[optional_input],
                            else_=['no_input']), '--dataset',
                        IfPresentPlaceholder(
                            input_name='optional_input', then=[dataset.uri], else_=['no_dataset'])
                    ],
                    args=['--output_path', output_path]
                )
    """

    def __init__(
        self,
        input_name: str,
        then: Union['CommandLineElement', List['CommandLineElement']],
        else_: Optional[Union['CommandLineElement',
                              List['CommandLineElement']]] = None,
    ) -> None:
        self.input_name = input_name
        self.then = then
        self.else_ = else_

    def _validate_then_and_else_are_only_single_element(self) -> None:
        """Rercursively validate that then and else contain only a single
        element.

        This method should only be called by a ConcatPlaceholder, which
        cannot have an IfPresentPlaceholder with a list in either 'then'
        or 'else_'.
        """

        # the illegal state
        if isinstance(self.then, list) or isinstance(self.else_, list):
            raise ValueError(
                f'Cannot use {IfPresentPlaceholder.__name__} within {ConcatPlaceholder.__name__} when `then` and `else_` arguments to {IfPresentPlaceholder.__name__} are lists. Please use a single element for `then` and `else_` only.'
            )

        # check that there is no illegal state found recursively
        if isinstance(self.then, ConcatPlaceholder):
            for item in self.then.items:
                if isinstance(item, IfPresentPlaceholder):
                    item._validate_then_and_else_are_only_single_element()
        elif isinstance(self.then, IfPresentPlaceholder):
            self.then._validate_then_and_else_are_only_single_element()

        if isinstance(self.else_, ConcatPlaceholder):
            for item in self.else_.items:
                if isinstance(item, IfPresentPlaceholder):
                    item._validate_then_and_else_are_only_single_element()
        elif isinstance(self.else_, IfPresentPlaceholder):
            self.else_._validate_then_and_else_are_only_single_element()

    def _to_dict(self) -> Dict[str, Any]:
        struct = {
            'IfPresent': {
                'InputName':
                    self.input_name,
                'Then': [
                    convert_command_line_element_to_string_or_struct(e)
                    for e in self.then
                ] if isinstance(self.then, list) else
                        convert_command_line_element_to_string_or_struct(
                            self.then)
            }
        }
        if self.else_:
            struct['IfPresent']['Else'] = [
                convert_command_line_element_to_string_or_struct(e)
                for e in self.else_
            ] if isinstance(
                self.else_,
                list) else convert_command_line_element_to_string_or_struct(
                    self.else_)
        return struct

    def _to_string(self) -> str:
        return json.dumps(self._to_dict())

    def __str__(self) -> str:
        raise ValueError(
            f'Cannot use {self.__class__.__name__} in an f-string.')


_CONTAINER_PLACEHOLDERS = (IfPresentPlaceholder, ConcatPlaceholder)
PRIMITIVE_INPUT_PLACEHOLDERS = (InputValuePlaceholder, InputPathPlaceholder,
                                InputUriPlaceholder, InputMetadataPlaceholder,
                                InputListOfArtifactsPlaceholder)
PRIMITIVE_OUTPUT_PLACEHOLDERS = (OutputParameterPlaceholder,
                                 OutputPathPlaceholder, OutputUriPlaceholder,
                                 OutputMetadataPlaceholder,
                                 OutputListOfArtifactsPlaceholder)

CommandLineElement = Union[str, Placeholder]


def convert_command_line_element_to_string(
        element: Union[str, Placeholder]) -> str:
    return element._to_string() if isinstance(element, Placeholder) else element


def convert_command_line_element_to_string_or_struct(
        element: Union[Placeholder, Any]) -> Any:
    if isinstance(element, Placeholder):
        return element._to_dict() if isinstance(
            element, _CONTAINER_PLACEHOLDERS) else element._to_string()

    return element


def maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
        arg: Dict[str, Any],
        component_dict: Dict[str, Any]) -> Union[CommandLineElement, Any]:
    if isinstance(arg, str):
        return arg

    if not isinstance(arg, dict):
        raise ValueError

    has_one_entry = len(arg) == 1

    if not has_one_entry:
        raise ValueError(
            f'Got unexpected dictionary {arg}. Expected a dictionary with one entry.'
        )

    first_key = list(arg.keys())[0]
    first_value = list(arg.values())[0]
    if first_key == 'inputValue':
        return InputValuePlaceholder(
            input_name=utils.sanitize_input_name(first_value))

    elif first_key == 'inputPath':
        return InputPathPlaceholder(
            input_name=utils.sanitize_input_name(first_value))

    elif first_key == 'inputUri':
        return InputUriPlaceholder(
            input_name=utils.sanitize_input_name(first_value))

    elif first_key == 'outputPath':
        outputs = component_dict['outputs']
        for output in outputs:
            if output['name'] == first_value:
                type_ = output.get('type')
                is_parameter = type_utils.is_parameter_type(type_)
                if is_parameter:
                    return OutputParameterPlaceholder(
                        output_name=utils.sanitize_input_name(first_value))
                else:
                    return OutputPathPlaceholder(
                        output_name=utils.sanitize_input_name(first_value))
        raise ValueError(
            f'{first_value} not found in component outputs. Could not process placeholders. Component spec: {component_dict}.'
        )

    elif first_key == 'outputUri':
        return OutputUriPlaceholder(
            output_name=utils.sanitize_input_name(first_value))

    elif first_key == 'ifPresent':
        structure_kwargs = arg['ifPresent']
        structure_kwargs['input_name'] = structure_kwargs.pop('inputName')
        structure_kwargs['otherwise'] = structure_kwargs.pop('else')
        structure_kwargs['then'] = [
            maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                e, component_dict=component_dict)
            for e in structure_kwargs['then']
        ]
        structure_kwargs['otherwise'] = [
            maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                e, component_dict=component_dict)
            for e in structure_kwargs['otherwise']
        ]
        return IfPresentPlaceholder(**structure_kwargs)

    elif first_key == 'concat':
        return ConcatPlaceholder(items=[
            maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                e, component_dict=component_dict) for e in arg['concat']
        ])

    elif first_key == 'executorInput':
        return ExecutorInputPlaceholder()

    elif 'if' in arg:
        if_ = arg['if']
        input_name = utils.sanitize_input_name(if_['cond']['isPresent'])
        then = if_['then']
        else_ = if_.get('else')

        if isinstance(then, list):
            then = [
                maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                    val, component_dict=component_dict) for val in then
            ]
        else:
            then = maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                then, component_dict=component_dict)

        if else_ is None:
            pass
        elif isinstance(else_, list):
            else_ = [
                maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                    val, component_dict=component_dict) for val in else_
            ]
        else:
            maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                else_, component_dict=component_dict)

        return IfPresentPlaceholder(
            input_name=input_name, then=then, else_=else_)

    elif 'concat' in arg:

        return ConcatPlaceholder(items=[
            maybe_convert_v1_yaml_placeholder_to_v2_placeholder(
                val, component_dict=component_dict) for val in arg['concat']
        ])
    else:
        raise TypeError(f'Unexpected argument {arg} of type {type(arg)}.')



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for v2-compatible compiled pipelines."""

import json
import os
import re
import tempfile
import unittest
from typing import Callable, Dict

import yaml
from kfp.deprecated import compiler
from kfp.deprecated import dsl as v1dsl
from kfp import dsl
from kfp.dsl import Artifact, InputPath, OutputPath, component


@component
def preprocess(uri: str, some_int: int, output_parameter_one: OutputPath(int),
               output_dataset_one: OutputPath('Dataset')):
    """Dummy Preprocess Step."""
    with open(output_dataset_one, 'w') as f:
        f.write('Output dataset')
    with open(output_parameter_one, 'w') as f:
        f.write("{}".format(1234))


@component
def train(dataset: InputPath('Dataset'),
          model: OutputPath('Model'),
          num_steps: int = 100):
    """Dummy Training Step."""

    with open(dataset, 'r') as input_file:
        input_string = input_file.read()
        with open(model, 'w') as output_file:
            for i in range(num_steps):
                output_file.write("Step {}\n{}\n=====\n".format(
                    i, input_string))


@unittest.skip('v2 compatible mode is being deprecated in SDK v2.0')
class TestV2CompatibleModeCompiler(unittest.TestCase):

    def setUp(self) -> None:
        self.maxDiff = None
        return super().setUp()

    def _ignore_kfp_version_in_template(self, template):
        """Ignores kfp sdk versioning in container spec."""

    def _assert_compiled_pipeline_equals_golden(self,
                                                kfp_compiler: compiler.Compiler,
                                                pipeline_func: Callable,
                                                golden_yaml_filename: str):
        compiled_file = os.path.join(tempfile.mkdtemp(), 'workflow.yaml')
        kfp_compiler.compile(pipeline_func, package_path=compiled_file)

        test_data_dir = os.path.join(os.path.dirname(__file__), 'testdata')
        golden_file = os.path.join(test_data_dir, golden_yaml_filename)

        def _load_compiled_template(filename: str) -> Dict:
            with open(filename, 'r') as f:
                workflow = yaml.safe_load(f)

            del workflow['metadata']
            for template in workflow['spec']['templates']:
                template.pop('metadata', None)

                if 'initContainers' not in template:
                    continue
                # Strip off the launcher image label before comparison
                for initContainer in template['initContainers']:
                    initContainer['image'] = initContainer['image'].split(
                        ':')[0]

                if 'container' in template:
                    template['container'] = json.loads(
                        re.sub("'kfp==(\d+).(\d+).(\d+)'", 'kfp',
                               json.dumps(template['container'])))

            return workflow

        golden = _load_compiled_template(golden_file)
        compiled = _load_compiled_template(compiled_file)

        # Devs can run the following command to update golden files:
        # UPDATE_GOLDENS=True python3 -m unittest kfp/compiler/v2_compatible_compiler_test.py
        # If UPDATE_GOLDENS=True, and the diff is
        # different, update the golden file and reload it.
        update_goldens = os.environ.get('UPDATE_GOLDENS', False)
        if golden != compiled and update_goldens:
            kfp_compiler.compile(pipeline_func, package_path=golden_file)
            golden = _load_compiled_template(golden_file)

        self.assertDictEqual(golden, compiled)

    def test_two_step_pipeline(self):

        @dsl.pipeline(
            pipeline_root='gs://output-directory/v2-artifacts',
            name='my-test-pipeline')
        def v2_compatible_two_step_pipeline():
            preprocess_task = preprocess(uri='uri-to-import', some_int=12)
            train_task = train(
                num_steps=preprocess_task.outputs['output_parameter_one'],
                dataset=preprocess_task.outputs['output_dataset_one'])

        kfp_compiler = compiler.Compiler(
            mode=v1dsl.PipelineExecutionMode.V2_COMPATIBLE)
        self._assert_compiled_pipeline_equals_golden(
            kfp_compiler, v2_compatible_two_step_pipeline,
            'v2_compatible_two_step_pipeline.yaml')

    def test_custom_launcher(self):

        @dsl.pipeline(
            pipeline_root='gs://output-directory/v2-artifacts',
            name='my-test-pipeline-with-custom-launcher')
        def v2_compatible_two_step_pipeline():
            preprocess_task = preprocess(uri='uri-to-import', some_int=12)
            train_task = train(
                num_steps=preprocess_task.outputs['output_parameter_one'],
                dataset=preprocess_task.outputs['output_dataset_one'])

        kfp_compiler = compiler.Compiler(
            mode=v1dsl.PipelineExecutionMode.V2_COMPATIBLE,
            launcher_image='my-custom-image')
        self._assert_compiled_pipeline_equals_golden(
            kfp_compiler, v2_compatible_two_step_pipeline,
            'v2_compatible_two_step_pipeline_with_custom_launcher.yaml')

    def test_constructing_container_op_directly_should_error(self):

        @dsl.pipeline(name='test-pipeline')
        def my_pipeline():
            v1dsl.ContainerOp(
                name='comp1',
                image='gcr.io/dummy',
                command=['python', 'main.py'])

        with self.assertRaisesRegex(
                RuntimeError,
                'Constructing ContainerOp instances directly is deprecated and not '
                'supported when compiling to v2 \(using v2 compiler or v1 compiler '
                'with V2_COMPATIBLE or V2_ENGINE mode\).'):
            compiler.Compiler(
                mode=v1dsl.PipelineExecutionMode.V2_COMPATIBLE).compile(
                    pipeline_func=my_pipeline, package_path='result.json')

    def test_use_importer_should_error(self):

        @dsl.pipeline(name='test-pipeline')
        def my_pipeline():
            dsl.importer(artifact_uri='dummy', artifact_class=Artifact)

        with self.assertRaisesRegex(
                ValueError,
                'dsl.importer is not supported with v1 compiler.',
        ):
            compiler.Compiler(
                mode=v1dsl.PipelineExecutionMode.V2_COMPATIBLE).compile(
                    pipeline_func=my_pipeline, package_path='result.json')


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional
import json
import click
from kfp.deprecated.cli.output import print_output, OutputFormat
import kfp_server_api


@click.group()
def recurring_run():
    """Manage recurring-run resources."""


@recurring_run.command()
@click.option(
    '--catchup/--no-catchup',
    help='Whether the recurring run should catch up if behind schedule.',
    type=bool)
@click.option(
    '--cron-expression',
    help='A cron expression representing a set of times, using 6 space-separated fields, e.g. "0 0 9 ? * 2-6".'
)
@click.option('--description', help='The description of the recurring run.')
@click.option(
    '--enable-caching/--disable-caching',
    help='Optional. Whether or not to enable caching for the run.',
    type=bool)
@click.option(
    '--enabled/--disabled',
    help='A bool indicating whether the recurring run is enabled or disabled.',
    type=bool)
@click.option(
    '--end-time',
    help='The RFC3339 time string of the time when to end the job.')
@click.option(
    '--experiment-id',
    help='The ID of the experiment to create the recurring run under, can only supply either an experiment ID or name.'
)
@click.option(
    '--experiment-name',
    help='The name of the experiment to create the recurring run under, can only supply either an experiment ID or name.'
)
@click.option('--job-name', help='The name of the recurring run.')
@click.option(
    '--interval-second',
    help='Integer indicating the seconds between two recurring runs in for a periodic schedule.'
)
@click.option(
    '--max-concurrency',
    help='Integer indicating how many jobs can be run in parallel.',
    type=int)
@click.option(
    '--pipeline-id',
    help='The ID of the pipeline to use to create the recurring run.')
@click.option(
    '--pipeline-package-path',
    help='Local path of the pipeline package(the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).'
)
@click.option(
    '--start-time',
    help='The RFC3339 time string of the time when to start the job.')
@click.option('--version-id', help='The id of a pipeline version.')
@click.argument("args", nargs=-1)
@click.pass_context
def create(ctx: click.Context,
           job_name: str,
           experiment_id: Optional[str] = None,
           experiment_name: Optional[str] = None,
           catchup: Optional[bool] = None,
           cron_expression: Optional[str] = None,
           enabled: Optional[bool] = None,
           description: Optional[str] = None,
           enable_caching: Optional[bool] = None,
           end_time: Optional[str] = None,
           interval_second: Optional[int] = None,
           max_concurrency: Optional[int] = None,
           params: Optional[dict] = None,
           pipeline_package_path: Optional[str] = None,
           pipeline_id: Optional[str] = None,
           start_time: Optional[str] = None,
           version_id: Optional[str] = None,
           args: Optional[List[str]] = None):
    """Create a recurring run."""
    client = ctx.obj["client"]
    output_format = ctx.obj['output']

    if (experiment_id is None) == (experiment_name is None):
        raise ValueError('Either experiment_id or experiment_name is required')
    if not experiment_id:
        experiment = client.create_experiment(experiment_name)
        experiment_id = experiment.id

    # Ensure we only split on the first equals char so the value can contain
    # equals signs too.
    split_args: List = [arg.split("=", 1) for arg in args or []]
    params: Dict[str, Any] = dict(split_args)
    recurring_run = client.create_recurring_run(
        cron_expression=cron_expression,
        description=description,
        enabled=enabled,
        enable_caching=enable_caching,
        end_time=end_time,
        experiment_id=experiment_id,
        interval_second=interval_second,
        job_name=job_name,
        max_concurrency=max_concurrency,
        no_catchup=not catchup,
        params=params,
        pipeline_package_path=pipeline_package_path,
        pipeline_id=pipeline_id,
        start_time=start_time,
        version_id=version_id)

    _display_recurring_run(recurring_run, output_format)


@recurring_run.command()
@click.option(
    '-e',
    '--experiment-id',
    help='Parent experiment ID of listed recurring runs.')
@click.option(
    '--page-token', default='', help="Token for starting of the page.")
@click.option(
    '-m',
    '--max-size',
    default=100,
    help="Max size of the listed recurring runs.")
@click.option(
    '--sort-by',
    default="created_at desc",
    help="Can be '[field_name]', '[field_name] desc'. For example, 'name desc'."
)
@click.option(
    '--filter',
    help=(
        "filter: A url-encoded, JSON-serialized Filter protocol buffer "
        "(see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto))."
    ))
@click.pass_context
def list(ctx: click.Context, experiment_id: str, page_token: str, max_size: int,
         sort_by: str, filter: str):
    """List recurring runs."""
    client = ctx.obj['client']
    output_format = ctx.obj['output']

    response = client.list_recurring_runs(
        experiment_id=experiment_id,
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    if response.jobs:
        _display_recurring_runs(response.jobs, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = "No recurring runs found"
        click.echo(msg)


@recurring_run.command()
@click.argument("job-id")
@click.pass_context
def get(ctx: click.Context, job_id: str):
    """Get detailed information about an experiment."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.get_recurring_run(job_id)
    _display_recurring_run(response, output_format)


@recurring_run.command()
@click.argument("job-id")
@click.pass_context
def delete(ctx: click.Context, job_id: str):
    """Delete a recurring run."""
    client = ctx.obj["client"]
    client.delete_job(job_id)


def _display_recurring_runs(recurring_runs: List[kfp_server_api.ApiJob],
                            output_format: OutputFormat):
    headers = ["Recurring Run ID", "Name"]
    data = [[rr.id, rr.name] for rr in recurring_runs]
    print_output(data, headers, output_format, table_format="grid")


def _display_recurring_run(recurring_run: kfp_server_api.ApiJob,
                           output_format: OutputFormat):
    table = [
        ["Recurring Run ID", recurring_run.id],
        ["Name", recurring_run.name],
    ]
    if output_format == OutputFormat.table.name:
        print_output([], ["Recurring Run Details"], output_format)
        print_output(table, [], output_format, table_format="plain")
    elif output_format == OutputFormat.json.name:
        print_output(dict(table), [], output_format)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper utils used in artifact and ontology_artifact classes."""

from typing import Any, Dict, Tuple

import enum
import jsonschema
import os
import yaml


class SchemaFieldType(enum.Enum):
    """Supported Schema field types."""
    NUMBER = 'number'
    INTEGER = 'integer'
    STRING = 'string'
    BOOL = 'bool'
    OBJECT = 'object'
    ARRAY = 'array'


def parse_schema(yaml_schema: str) -> Tuple[str, Dict[str, SchemaFieldType]]:
    """Parses yaml schema.

    Ensures that schema is well-formed and returns dictionary of properties and
    its type for type-checking.

    Args:
      yaml_schema: Yaml schema to be parsed.

    Returns:
      str: Title set in the schema.
      Dict: Property name to SchemaFieldType enum.

    Raises:
      ValueError if title field is not set in schema or an
        unsupported(i.e. not defined in SchemaFieldType)
        type is specified for the field.
    """

    schema = yaml.full_load(yaml_schema)
    if 'title' not in schema.keys():
        raise ValueError('Invalid _schema, title must be set. \
      Got: {}'.format(yaml_schema))

    title = schema['title']
    properties = {}
    if 'properties' in schema.keys():
        schema_properties = schema['properties'] or {}
        for property_name, property_def in schema_properties.items():
            try:
                properties[property_name] = SchemaFieldType(
                    property_def['type'])
            except ValueError:
                raise ValueError('Unsupported type:{} specified for field: {} \
          in schema'.format(property_def['type'], property_name))

    return title, properties


def verify_schema_instance(schema: str, instance: Dict[str, Any]):
    """Verifies instnace is well-formed against the schema.

    Args:
      schema: Schema to use for verification.
      instance: Object represented as Dict to be verified.

    Raises:
      RuntimeError if schema is not well-formed or instance is invalid against
       the schema.
    """

    if len(instance) == 0:
        return

    try:
        jsonschema.validate(instance=instance, schema=yaml.full_load(schema))
    except jsonschema.exceptions.SchemaError:
        raise RuntimeError('Invalid schema schema: {} used for \
        verification'.format(schema))
    except jsonschema.exceptions.ValidationError:
        raise RuntimeError('Invalid values set: {} in object for schema: \
      {}'.format(instance, schema))


def read_schema_file(schema_file: str) -> str:
    """Reads yamls schema from type_scheams folder.

    Args:
      schema_file: Name of the file to read schema from.

    Returns:
      Read schema from the schema file.
    """
    schema_file_path = os.path.join(
        os.path.dirname(__file__), 'type_schemas', schema_file)

    with open(schema_file_path) as schema_file:
        return schema_file.read()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from kfp import client
from kfp.cli.utils import parsing


class MyClass:
    """Class init.

    Args:
        a (str): String a.
        b (str): String b.
    """

    def __init__(self, a: str, b: str):
        """Class init.

        Args:
            a (str): String a.
            b (str): String b.
        """
        self.a = a
        self.b = b

    def method(self, a: str, b: str):
        """My method.

        Args:
            a (str): String a.
            b (str): String b.

        Returns:
            str: Concat str.
        """
        return a + b


class TestGetParamDescr(unittest.TestCase):

    def test_function(self):

        def fn(a: str, b: str):
            """My fn.

            Args:
                a (str): String a.
                b (str): String b.

            Returns:
                str: Concat str.
            """
            return a + b

        a_descr = parsing.get_param_descr(fn, 'a')
        self.assertEqual(a_descr, 'String a.')
        b_descr = parsing.get_param_descr(fn, 'b')
        self.assertEqual(b_descr, 'String b.')

    def test_function_without_type_info(self):

        def fn(a: str, b: str):
            """My fn.

            Args:
                a: String a.
                b: String b.

            Returns:
                str: Concat str.
            """
            return a + b

        a_descr = parsing.get_param_descr(fn, 'a')
        self.assertEqual(a_descr, 'String a.')
        b_descr = parsing.get_param_descr(fn, 'b')
        self.assertEqual(b_descr, 'String b.')

    def test_class(self):

        class_doc_a = parsing.get_param_descr(MyClass, 'a')
        self.assertEqual(class_doc_a, 'String a.')
        class_doc_b = parsing.get_param_descr(MyClass, 'b')
        self.assertEqual(class_doc_b, 'String b.')

        init_a = parsing.get_param_descr(MyClass.__init__, 'a')
        self.assertEqual(init_a, 'String a.')
        init_b = parsing.get_param_descr(MyClass.__init__, 'b')
        self.assertEqual(init_b, 'String b.')

        method_a = parsing.get_param_descr(MyClass.method, 'a')
        self.assertEqual(method_a, 'String a.')
        method_b = parsing.get_param_descr(MyClass.method, 'b')
        self.assertEqual(method_b, 'String b.')

    def test_instance(self):
        instance = MyClass('a', 'b')

        inst_doc_a = parsing.get_param_descr(instance, 'a')
        self.assertEqual(inst_doc_a, 'String a.')
        inst_doc_b = parsing.get_param_descr(instance, 'b')
        self.assertEqual(inst_doc_b, 'String b.')

        init_a = parsing.get_param_descr(instance.__init__, 'a')
        self.assertEqual(init_a, 'String a.')
        init_b = parsing.get_param_descr(instance.__init__, 'b')
        self.assertEqual(init_b, 'String b.')

        method_a = parsing.get_param_descr(instance.method, 'a')
        self.assertEqual(method_a, 'String a.')
        method_b = parsing.get_param_descr(instance.method, 'b')
        self.assertEqual(method_b, 'String b.')

    def test_multiline(self):
        host_descr = parsing.get_param_descr(client.Client, 'host')

        self.assertEqual(
            host_descr,
            "Host name to use to talk to Kubeflow Pipelines. If not set, the in-cluster service DNS name will be used, which only works if the current environment is a pod in the same cluster (such as a Jupyter instance spawned by Kubeflow's JupyterHub). (`More information on connecting. <https://www.kubeflow.org/docs/components/pipelines/sdk/connect-api/>`_)"
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import configparser
import pathlib
from typing import Dict, Optional
import warnings

_KFP_CONFIG_FILE = 'kfp_config.ini'

_COMPONENTS_SECTION = 'Components'


class KFPConfig():
    """Class for managing KFP component configuration.

    The configuration is .ini file named `kfp_config.ini` that can be parsed by
    Python's native configparser module. Currently, this class supports a single
    `Components` section, which lists components as key-value pairs. The key is
    the component name (i.e. the function name), and the value is the path to
    the file containing this function. The path is usually relative from the
    location of the configuration file, but absolute paths should also work.

    At runtime, the KFP v2 Executor, defined in executor_main.py, will look
    for this configuration file in its current working directory. If found,
    it will load its contents, and use this to find the file containing the
    component to execute.

    Example of the file's contents:

    [Components]
    my_component_1 = my_dir_1/my_component_1.py
    my_component_2 = my_dir_2/my_component_2.py
    ...
    """

    def __init__(self, config_directory: Optional[pathlib.Path] = None):
        """Creates a KFPConfig object.

        Loads the config from an existing `kfp_config.ini` file if found.

        Args:
            config_directory: Looks for a file named `kfp_config.ini` in this
                directory. Defaults to the current directory.
        """
        self._config_parser = configparser.ConfigParser()
        # Preserve case for keys.
        self._config_parser.optionxform = lambda x: x

        if config_directory is None:
            self._config_filepath = pathlib.Path(_KFP_CONFIG_FILE)
        else:
            self._config_filepath = config_directory / _KFP_CONFIG_FILE

        try:
            with open(str(self._config_filepath), 'r') as f:
                self._config_parser.read_file(f)
        except IOError:
            warnings.warn('No existing KFP Config file found')

        if not self._config_parser.has_section(_COMPONENTS_SECTION):
            self._config_parser.add_section(_COMPONENTS_SECTION)

        self._components = {}

    def add_component(self, function_name: str, path: pathlib.Path):
        """Adds a KFP component.

        Args:
            function_name: The name of the component function.
            path: A path to the file containing the component.
        """
        self._components[function_name] = str(path)

    def save(self):
        """Writes out a KFP config file."""
        # Always write out components in alphabetical order for determinism,
        # especially in tests.
        for function_name in sorted(self._components.keys()):
            self._config_parser[_COMPONENTS_SECTION][
                function_name] = self._components[function_name]

        with open(str(self._config_filepath), 'w') as f:
            self._config_parser.write(f)

    def get_components(self) -> Dict[str, pathlib.Path]:
        """Returns a list of known KFP components.

        Returns:
            A dictionary from component name (function name) to a pathlib.Path
            pointing to the Python file with this component's definition.
        """
        return {
            function_name: pathlib.Path(module_path) for function_name,
            module_path in self._config_parser[_COMPONENTS_SECTION].items()
        }



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pipeline as a component (aka graph component)."""

import inspect
from typing import Callable, Optional
import uuid

from kfp.compiler import pipeline_spec_builder as builder
from kfp.components import base_component
from kfp.components import pipeline_channel
from kfp.components import pipeline_context
from kfp.components import structures
from kfp.pipeline_spec import pipeline_spec_pb2


class GraphComponent(base_component.BaseComponent):
    """A component defined via @dsl.pipeline decorator.

    Attribute:
        pipeline_func: The function that becomes the implementation of this component.
    """

    def __init__(
        self,
        component_spec: structures.ComponentSpec,
        pipeline_func: Callable,
        display_name: Optional[str] = None,
    ):
        super().__init__(component_spec=component_spec)
        self.pipeline_func = pipeline_func

        args_list = []
        signature = inspect.signature(pipeline_func)

        for arg_name in signature.parameters:
            input_spec = component_spec.inputs[arg_name]
            args_list.append(
                pipeline_channel.create_pipeline_channel(
                    name=arg_name,
                    channel_type=input_spec.type,
                    is_artifact_list=input_spec.is_artifact_list,
                ))

        with pipeline_context.Pipeline(
                self.component_spec.name) as dsl_pipeline:
            pipeline_outputs = pipeline_func(*args_list)

        if not dsl_pipeline.tasks:
            raise ValueError('Task is missing from pipeline.')

        # Making the pipeline group name unique to prevent name clashes with
        # templates
        pipeline_group = dsl_pipeline.groups[0]
        pipeline_group.name = uuid.uuid4().hex

        pipeline_spec, platform_spec = builder.create_pipeline_spec(
            pipeline=dsl_pipeline,
            component_spec=self.component_spec,
            pipeline_outputs=pipeline_outputs,
        )

        pipeline_root = getattr(pipeline_func, 'pipeline_root', None)
        if pipeline_root is not None:
            pipeline_spec.default_pipeline_root = pipeline_root
        if display_name is not None:
            pipeline_spec.pipeline_info.display_name = display_name
        if component_spec.description is not None:
            pipeline_spec.pipeline_info.description = component_spec.description

        self.component_spec.implementation.graph = pipeline_spec
        self.component_spec.platform_spec = platform_spec

    @property
    def pipeline_spec(self) -> pipeline_spec_pb2.PipelineSpec:
        """Returns the pipeline spec of the component."""
        return self.component_spec.implementation.graph

    def execute(self, **kwargs):
        raise RuntimeError('Graph component has no local execution mode.')



# Copyright 2018-2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union

from kfp.deprecated.dsl import _for_loop, _pipeline_param

from . import _container_op
from . import _pipeline


class OpsGroup(object):
    """Represents a logical group of ops and group of OpsGroups.

    This class is the base class for groups of ops, such as ops sharing
    an exit handler, a condition branch, or a loop. This class is not
    supposed to be used by pipeline authors. It is useful for
    implementing a compiler.
    """

    def __init__(self,
                 group_type: str,
                 name: str = None,
                 parallelism: int = None):
        """Create a new instance of OpsGroup.

        Args:
          group_type (str): one of 'pipeline', 'exit_handler', 'condition',
            'for_loop', and 'graph'.
          name (str): name of the opsgroup
          parallelism (int): parallelism for the sub-DAG:s
        """
        #TODO: declare the group_type to be strongly typed
        self.type = group_type
        self.ops = list()
        self.groups = list()
        self.name = name
        self.dependencies = []
        self.parallelism = parallelism
        # recursive_ref points to the opsgroups with the same name if exists.
        self.recursive_ref = None

        self.loop_args = None

    @staticmethod
    def _get_matching_opsgroup_already_in_pipeline(group_type, name):
        """Retrieves the opsgroup when the pipeline already contains it.

        the opsgroup might be already in the pipeline in case of recursive calls.

        Args:
          group_type (str): one of 'pipeline', 'exit_handler', 'condition', and
            'graph'.
          name (str): the name before conversion.
        """
        if not _pipeline.Pipeline.get_default_pipeline():
            raise ValueError('Default pipeline not defined.')
        if name is None:
            return None
        name_pattern = '^' + (group_type + '-' + name + '-').replace(
            '_', '-') + '[\d]+$'
        for ops_group_already_in_pipeline in _pipeline.Pipeline.get_default_pipeline(
        ).groups:
            import re
            if ops_group_already_in_pipeline.type == group_type \
                    and re.match(name_pattern ,ops_group_already_in_pipeline.name):
                return ops_group_already_in_pipeline
        return None

    def _make_name_unique(self):
        """Generate a unique opsgroup name in the pipeline."""
        if not _pipeline.Pipeline.get_default_pipeline():
            raise ValueError('Default pipeline not defined.')

        self.name = (
            self.type + '-' + ('' if self.name is None else self.name + '-') +
            str(_pipeline.Pipeline.get_default_pipeline().get_next_group_id()))
        self.name = self.name.replace('_', '-')

    def __enter__(self):
        if not _pipeline.Pipeline.get_default_pipeline():
            raise ValueError('Default pipeline not defined.')

        self.recursive_ref = self._get_matching_opsgroup_already_in_pipeline(
            self.type, self.name)
        if not self.recursive_ref:
            self._make_name_unique()

        _pipeline.Pipeline.get_default_pipeline().push_ops_group(self)
        return self

    def __exit__(self, *args):
        _pipeline.Pipeline.get_default_pipeline().pop_ops_group()

    def after(self, *ops):
        """Specify explicit dependency on other ops."""
        for op in ops:
            self.dependencies.append(op)
        return self

    def remove_op_recursive(self, op):
        if self.ops and op in self.ops:
            self.ops.remove(op)
        for sub_group in self.groups or []:
            sub_group.remove_op_recursive(op)


class SubGraph(OpsGroup):
    TYPE_NAME = 'subgraph'

    def __init__(self, parallelism: int):
        if parallelism < 1:
            raise ValueError(
                'SubGraph parallism set to < 1, allowed values are > 0')
        super(SubGraph, self).__init__(self.TYPE_NAME, parallelism=parallelism)


class ExitHandler(OpsGroup):
    """Represents an exit handler that is invoked upon exiting a group of ops.

    Args:
      exit_op: An operator invoked at exiting a group of ops.

    Raises:
      ValueError: Raised if the exit_op is invalid.

    Example:
      ::

        exit_op = ContainerOp(...)
        with ExitHandler(exit_op):
          op1 = ContainerOp(...)
          op2 = ContainerOp(...)
    """

    def __init__(self, exit_op: _container_op.ContainerOp):
        super(ExitHandler, self).__init__('exit_handler')
        if exit_op.dependent_names:
            raise ValueError('exit_op cannot depend on any other ops.')

        # Removing exit_op form any group
        _pipeline.Pipeline.get_default_pipeline().remove_op_from_groups(exit_op)

        # Setting is_exit_handler since the compiler might be using this attribute.
        # TODO: Check that it's needed
        exit_op.is_exit_handler = True

        self.exit_op = exit_op


class Condition(OpsGroup):
    """Represents an condition group with a condition.

    Args:
      condition (ConditionOperator): the condition.
      name (str): name of the condition
    Example: ::
        with Condition(param1=='pizza', '[param1 is pizza]'): op1 =
          ContainerOp(...) op2 = ContainerOp(...)
    """

    def __init__(self, condition, name=None):
        super(Condition, self).__init__('condition', name)
        self.condition = condition


class Graph(OpsGroup):
    """Graph DAG with inputs, recursive_inputs, and outputs.

    This is not used directly by the users but auto generated when the
    graph_component decoration exists

    Args:
      name: Name of the graph.
    """

    def __init__(self, name):
        super(Graph, self).__init__(group_type='graph', name=name)
        self.inputs = []
        self.outputs = {}
        self.dependencies = []


class ParallelFor(OpsGroup):
    """Represents a parallel for loop over a static set of items.

    Example:
      In this case :code:`op1` would be executed twice, once with case
      :code:`args=['echo 1']` and once with case :code:`args=['echo 2']`::

        with dsl.ParallelFor([{'a': 1, 'b': 10}, {'a': 2, 'b': 20}]) as item:
          op1 = ContainerOp(..., args=['echo {}'.format(item.a)])
          op2 = ContainerOp(..., args=['echo {}'.format(item.b])
    """
    TYPE_NAME = 'for_loop'

    def __init__(self,
                 loop_args: Union[_for_loop.ItemList,
                                  _pipeline_param.PipelineParam],
                 parallelism: int = None):
        if parallelism and parallelism < 1:
            raise ValueError(
                'ParallelFor parallism set to < 1, allowed values are > 0')

        self.items_is_pipeline_param = isinstance(loop_args,
                                                  _pipeline_param.PipelineParam)

        super().__init__(self.TYPE_NAME, parallelism=parallelism)

        if self.items_is_pipeline_param:
            loop_args = _for_loop.LoopArguments.from_pipeline_param(loop_args)
        elif not self.items_is_pipeline_param and not isinstance(
                loop_args, _for_loop.LoopArguments):
            # we were passed a raw list, wrap it in loop args
            loop_args = _for_loop.LoopArguments(
                loop_args,
                code=str(_pipeline.Pipeline.get_default_pipeline()
                         .get_next_group_id()),
            )

        self.loop_args = loop_args

    def __enter__(self) -> _for_loop.LoopArguments:
        _ = super().__enter__()
        return self.loop_args
